import copy
import warnings

import numpy as np

from .rt_enum import FILTERED_LONG_NAME, INVALID_DICT, NumpyCharTypes, TypeRegister
from .rt_numpy import hstack, lexsort, mask_andi


class GroupByKeys:
    """Handles masking, appending invalid, and sorting of key columns for a groupby operation.

    Parameters
    ----------
    grouping_dict : dict
        Non-unique or unique key columns.
    ifirstkey : array, optional
        If set, first occurence to generate unique values from non-unique keys.
        Sometimes these are lazily evaluated and do not correspond to the grouping dict held.
        See the `prebinned` keyword.
    isortrows : array, optional
        A sorted index for the unique array.
        May be calculated later on, after `grouping_dict` is reduced to unique values.
    sort_display : bool, default False
        If True, unique keys in result of operation will be sorted. Otherwise, will appear unsorted.
    pre_sorted : bool, default False
        Unique `grouping_dict` is already in a sorted order, do not apply / calculate `isortrows`,
        even if sort on is True.
    prebinned : bool, default False
        If True, `grouping_dict` contains unique values. `ifirstkey` will not be stored.
        If False, `grouping_dict` contains non-unique values, the default if constructed by a `GroupBy` object.

    Notes
    -----
    **Constructor**

    GroupByKeys has two main ways of initialization:

    1. From a non-unique grouping_dict and ifirstkey (fancy index to unique values). The object
       will hold on to both, and lazily generate the groupby keys as necessary.
    2. From already binned gbkeys (unique values). Most categoricals will initialize GroupByKeys this way.
       Because Categoricals are sometimes naturally sorted, they may set the pre_sorted keyword to True.

    If sort_display is True and the keys are not already sorted, the gbkeys will be sorted in-place the first time
    a groupby calculation is made. After being sorted, the internal _sort_applied flag will be set. Despite the keys
    being sorted, the sort might still need to be applied to the data columns of the groupby calculation's result.

    **Lazy Evaluation**

    * If isortrows is not provided and the gbkeys are not pre-sorted, a lexsort will be performed, and the keys will be sorted inplace.
    * If gbkeys are requested with a filter bin, a new bin will be permanently prepended to each key array. After the filtered bin is added,
      the gbkeys will still default to return arrays without the filter (a view of the held arrays offset by 1).
    * Multikey labels are a list of strings of the tuples that will appear when a multikey is displayed. They will also add a filtered tuple
      as necessary, and default to a reduced view after the addition - just like the gbkeys. Multikey labels will not be generated until
      they are requested for display (because they are constructed in a python loop, generating these is expensive).
    """

    def __init__(
        self, grouping_dict, ifirstkey=None, isortrows=None, sort_display=False, pre_sorted=False, prebinned=False
    ):
        self._grouping_dict = grouping_dict
        self._gbkeys = None

        # arrays in the dict are already unique
        if prebinned:
            # iFirstKey is only relevant to non-unique arrays, discard it
            # will still be held in the Grouping object
            self._ifirstkey = None
            self._gbkeys = grouping_dict.copy()
            self._unique_count = len(list(grouping_dict.values())[0])

        else:
            self._ifirstkey = ifirstkey
            self._unique_count = len(ifirstkey)

        self._isortrows = isortrows
        self._sort_display = sort_display
        self._pre_sorted = pre_sorted

        self._sort_applied = False  # sort has been applied to the key columns
        self._filter_applied = False  # filter bin has been added to the key columns
        self._filter_applied_labels = False  # filter has been added to the multikey labels
        self._multikey_labels = None  # (lazily generated)

    @property
    # ------------------------------------------------------------
    def sort_gb_data(self):
        """
        If a sort has been applied to the gbkeys, they do not need to be sorted, however
        the data resulting from a groupby calculation is naturally unsorted and will still need
        a sort applied.
        """
        return self._sort_display and self._pre_sorted is False

    # ------------------------------------------------------------
    @property
    def isortrows(self):
        """
        Generates isortrows (index to sort groupby keys). Possibly performs a lexsort.
        """
        if self._isortrows is None:
            self._isortrows = self._make_isortrows()
        return self._isortrows

    # ------------------------------------------------------------
    @property
    def gbkeys(self):
        """
        Generates groupby keys if necessary. Returns groupby keys.
        """
        return self.keys()

    # ------------------------------------------------------------
    @property
    def gbkeys_filtered(self):
        """
        Adds a filter to the gbkeys, or returns the already filtered gbkeys.
        """
        self._insert_filter_bin()
        return self.keys(showfilter=True)

    # ------------------------------------------------------------
    def _insert_filter_bin(self):
        if self._filter_applied is False:
            # use the property for the iterator, so they can be generated if necessary
            for k, v in self.gbkeys.items():
                filtername = self._get_filter_bin_name(v)
                self._gbkeys[k] = hstack((filtername, v))
            self._filter_applied = True

    # ------------------------------------------------------------
    def unique_unsorted(self):
        """Pull the unique keys unsorted, using iFirstKey or the prebinned uniques."""
        if self._gbkeys is not None:
            return self._gbkeys
        else:
            return self._pull_from_ifirstkey()

    # ------------------------------------------------------------
    def _pull_from_ifirstkey(self):
        if self._gbkeys is None:
            if self._ifirstkey is not None:
                unique_keys = {}
                for k, v in self._grouping_dict.items():
                    unique_keys[k] = v[self._ifirstkey]
            else:
                raise ValueError(
                    f"Groupby keys need to be generated with an iFirstKey array, which the groupby key object does not have."
                )
        else:
            unique_keys = self._gbkeys
        return unique_keys

    # ------------------------------------------------------------
    def keys(self, sort=None, showfilter=False):
        """Return unique keys, possibly apply a sort and add a filter bin."""
        # groupby keys need to be generated from the non-unique grouping dict
        self._gbkeys = self._pull_from_ifirstkey()

        if sort is not None:
            self._sort_display = sort

        keys = self._gbkeys
        # only apply sort if requested and not already sorted
        if self._sort_display and self._pre_sorted is False:
            if self._sort_applied is False:
                if self._isortrows is None:
                    self._isortrows = self._make_isortrows()
                keys = {}
                for k, v in self._gbkeys.items():
                    keys[k] = v[self._isortrows]
                self._gbkeys = keys
                self._sort_applied = True

        # hide the filter bin by slicing the array if filter was turned off after adding
        if showfilter is False and self._filter_applied is True:
            keys = self._trim_keys(keys)

        return keys

    # ------------------------------------------------------------
    @property
    def multikey_labels(self):
        return self.labels()

    # ------------------------------------------------------------
    @property
    def multikey_labels_filtered(self):
        return self.labels(showfilter=True)

    # ------------------------------------------------------------
    def _insert_filter_label(self):
        labels = self.multikey_labels
        if self._filter_applied_labels is False:
            label = (FILTERED_LONG_NAME,) * len(self.gbkeys)
            self._multikey_labels.insert(0, label)
            self._filter_applied_labels = True

    # ------------------------------------------------------------
    def labels(self, showfilter=False):
        """
        Generates list of tuples from multikey columns.
        """
        if self._multikey_labels is None:
            self._multikey_labels = []
            for i in range(self.unique_count):
                key = self.get_bin_from_index(i)
                key = str(key)
                key.replace("'", "")
                self._multikey_labels.append(key)

        if self._filter_applied_labels and showfilter is False:
            return self._trim_keys(self._multikey_labels)
        return self._multikey_labels

    # ------------------------------------------------------------
    def _trim_keys(self, keys):
        """
        Return a trimmed view of the keys so the filtered bin is not included.
        Also trims list of multikey labels
        """

        if isinstance(keys, dict):
            trimmed = {}
            for k, v in keys.items():
                trimmed[k] = v[1:]
        else:
            trimmed = keys[1:]
        return trimmed

    # ------------------------------------------------------------
    def _get_filter_bin_name(self, arr):
        if arr.dtype.char == "U":
            filtername = FILTERED_LONG_NAME
        elif arr.dtype.char == "S":
            filtername = FILTERED_LONG_NAME.encode()
        else:
            # subclasses will use correct invalid and class will be preserved in hstack
            filtername = arr.fill_invalid(shape=1, inplace=False)
        return filtername

    # ------------------------------------------------------------
    @property
    def unique_count(self):
        """
        Returns number of unique groupby keys - lazily evaluated and stored.
        """
        # print('unique count hit!')
        if self._unique_count is None:
            first_key = list(self.gbkeys.values())[0]
            self._unique_count = len(first_key)
        return self._unique_count

    # ------------------------------------------------------------
    def _get_index_from_tuple(self, tup):
        """
        If the GroupByKeys object is holding a multikey dictionary, it can be indexed by
        a tuple. This internal routine (called by get_index_from_bin/__getitem__) will return
        the bin index of matching multikey entries or -1 if not found. Any string/bytes values
        will be fixed to match the string/bytes column.
        """
        if len(tup) == len(self.gbkeys):
            # build a boolean mask for each item in tuple
            match = []
            dictlist = list(self.gbkeys.values())
            for colnum, item in enumerate(tup):
                # match string types if necessary
                if isinstance(item, bytes):
                    if dictlist[colnum].dtype.char == "U":
                        item = bytes.decode(item)
                elif isinstance(item, str):
                    if dictlist[colnum].dtype.char == "S":
                        item = item.encode()
                match.append(dictlist[colnum] == item)
            match = mask_andi(match)
            match = np.where(match)[0]
            if len(match) == 1:
                return match[0]
            else:
                return -1
        else:
            raise ValueError(f"Tuple must contain an item for each groupby key column.")

    # ------------------------------------------------------------
    def get_index_from_bin(self, bin):
        """
        :param bin: a tuple of multiple keys or a single key (will be converted to tuple)

        :return index: the bin index, or -1 if not found.
        """
        if not isinstance(bin, tuple):
            bin = (bin,)
        return self._get_index_from_tuple(bin)

    # ------------------------------------------------------------
    def get_bin_from_index(self, index):
        """
        :param index: int or list of integers
        :return result_bins: matching bins for provided indices or an empty list
        """
        if index >= 0 and index < self.unique_count:
            result_bins = []
            for column in self.gbkeys.values():
                bin = column[index]
                if isinstance(bin, bytes):
                    bin = bytes.decode(bin)
                result_bins.append(bin)
            if len(result_bins) == 1:
                return result_bins[0]
            else:
                return tuple(result_bins)
        else:
            raise ValueError(f"Bin index {index} was out of range for gbkeys of length {self.unique_count}")

    # ------------------------------------------------------------
    def get_bin(self, index):
        """
        :param index: int or list of integers
        :return result_bins: matching bins for provided indices or an empty list
        """
        if isinstance(index, (int, np.integer)):
            result = self.get_bin_from_index(index)

        elif isinstance(index, list):
            if isinstance(index[0], (int, np.integer)):
                result = []
                for i in index:
                    result.append(self.get_bin_from_index(i))
                if len(result) == 1:
                    result = result[0]
            else:
                raise TypeError(f"Bins can only be retrieved by lists of integers, not {type(index[0])}")

        elif isinstance(index, np.ndarray):
            if index.dtype.char in NumpyCharTypes.AllInteger:
                result = []
                for i in index:
                    result.append(self.get_bin_from_index(i))
                if len(result) == 1:
                    result = result[0]
            else:
                raise TypeError(f"Bins can only be retrieved by numpy arrays of integer type, not {index.dtype}")

        else:
            raise TypeError(f"Bins must be retrieved by integer, not {type(index)}")

        return result

    # ------------------------------------------------------------
    def __getitem__(self, index):
        return self.get_bin(index)

    # ------------------------------------------------------------
    def _make_isortrows(self):
        sortlist = list(self._gbkeys.values())
        sortlist.reverse()
        isortrows = lexsort(sortlist)
        return isortrows

    # ------------------------------------------------------------
    def unsort(self):
        """
        Sets the internal _sort_display flag to False. Will warn the user if the groupby keys are already sorted or were pre-sorted
        when GroupByKeys were constructed.
        """
        if self._pre_sorted:
            warnings.warn("Groupby keys were naturally sorted.")
        elif self._sort_applied:
            raise ValueError("Groupby keys were already sorted. Data mis-match will occur.")
        self._sort_display = False

    # ------------------------------------------------------------
    def copy(self, deep=False):
        """
        Creates a deep or shallow copy of the grouping
        """
        if deep:
            new_keychain = GroupByKeys(copy.deepcopy(self._grouping_dict), ifirstkey=self._ifirstkey)
            if self._isortrows is not None:
                new_keychain._isortrows = self._isortrows.copy()

        else:
            new_keychain = GroupByKeys(self._grouping_dict.copy(), ifirstkey=self._ifirstkey)
            if self._isortrows is not None:
                new_keychain._isortrows = self._isortrows

        new_keychain._sort_display = self._sort_display
        new_keychain._pre_sorted = self._pre_sorted
        new_keychain._sort_applied = self._sort_applied
        new_keychain._filter_applied = self._filter_applied
        new_keychain._filter_applied_labels = self._filter_applied_labels
        new_keychain._unique_count = self._unique_count
        new_keychain._multikey_labels = self._multikey_labels

        return new_keychain

    # ------------------------------------------------------------
    def __repr__(self):
        return self._build_string()

    # ------------------------------------------------------------
    def __str__(self):
        return self._build_string()

    # ------------------------------------------------------------
    def _build_string(self):
        summary_str = []
        summary_str.append("gbkeys: " + str(self._gbkeys))
        summary_str.append("isortrows: " + str(self._isortrows))
        summary_str.append("sort on: " + str(self._sort_display))
        summary_str.append("naturally sorted: " + str(self._pre_sorted))
        summary_str.append("sort applied: " + str(self._sort_applied))
        summary_str.append("filter applied: " + str(self._filter_applied))
        return "\n".join(summary_str)

    # ------------------------------------------------------------
    @property
    def multikey(self):
        """
        Returns True if GroupByKeys object is holding multiple columns in _gbkeys
        """
        if self._gbkeys is not None:
            return len(self._gbkeys) > 1
        return False

    # ------------------------------------------------------------
    @property
    def singlekey(self):
        """
        Returns True if GroupByKeys object is holding a single column in _gbkeys
        """
        if self._gbkeys is not None:
            return len(self._gbkeys) == 1
        return False



__all__ = ["GroupBy"]

import warnings
from enum import IntEnum

import numpy as np
import riptide_cpp as rc

from .rt_enum import GB_FUNC_COUNT, GB_FUNCTIONS, GB_STRING_ALLOWED, CategoryMode
from .rt_groupbykeys import GroupByKeys
from .rt_groupbyops import GroupByOps
from .rt_numpy import (
    _groupbycalculateall,
    _groupbycalculateallpack,
    empty_like,
    groupbypack,
    sortinplaceindirect,
)
from .rt_timers import GetTSC
from .rt_utils import bool_to_fancy, mbget

#'''
# Series.abs()	Return an object with absolute value taken�only applicable to objects that are all numeric.
# Series.all([axis, bool_only, skipna, level])	Return whether all elements are True over requested axis
# Series.any([axis, bool_only, skipna, level])	Return whether any element is True over requested axis
# Series.autocorr([lag])	Lag-N autocorrelation
# Series.between(left, right[, inclusive])	Return boolean Series equivalent to left <= series <= right.
# Series.clip([lower, upper, axis, inplace])	Trim values at input threshold(s).
# Series.clip_lower(threshold[, axis, inplace])	Return copy of the input with values below given value(s) truncated.
# Series.clip_upper(threshold[, axis, inplace])	Return copy of input with values above given value(s) truncated.
# Series.corr(other[, method, min_periods])	Compute correlation with other Series, excluding missing values
# Series.count([level])	Return number of non-NA/null observations in the Series
# Series.cov(other[, min_periods])	Compute covariance with Series, excluding missing values
# Series.cummax([axis, skipna])	Return cumulative max over requested axis.
# Series.cummin([axis, skipna])	Return cumulative minimum over requested axis.
# Series.cumprod([axis, skipna])	Return cumulative product over requested axis.
# Series.cumsum([axis, skipna])	Return cumulative sum over requested axis.
# Series.describe([percentiles, include, exclude])	Generates descriptive statistics that summarize the central tendency, dispersion and shape of a datasets distribution, excluding NaN values.
# Series.diff([periods])	1st discrete difference of object
# Series.factorize([sort, na_sentinel])	Encode the object as an enumerated type or categorical variable
# Series.kurt([axis, skipna, level, numeric_only])	Return unbiased kurtosis over requested axis using Fisher�s definition of kurtosis (kurtosis of normal == 0.0).
# Series.mad([axis, skipna, level])	Return the mean absolute deviation of the values for the requested axis
# Series.max([axis, skipna, level, numeric_only])	This method returns the maximum of the values in the object.
# Series.mean([axis, skipna, level, numeric_only])	Return the mean of the values for the requested axis
# Series.median([axis, skipna, level, ...])	Return the median of the values for the requested axis
# Series.min([axis, skipna, level, numeric_only])	This method returns the minimum of the values in the object.
# Series.mode()	Return the mode(s) of the dataset.
# Series.nlargest([n, keep])	Return the largest n elements.
# Series.nsmallest([n, keep])	Return the smallest n elements.
# Series.pct_change([periods, fill_method, ...])	Percent change over given number of periods.
# Series.prod([axis, skipna, level, ...])	Return the product of the values for the requested axis
# Series.quantile([q, interpolation])	Return value at the given quantile, a la numpy.percentile.
# Series.rank([axis, method, numeric_only, ...])	Compute numerical data ranks (1 through n) along axis.
# Series.sem([axis, skipna, level, ddof, ...])	Return unbiased standard error of the mean over requested axis.
# Series.skew([axis, skipna, level, numeric_only])	Return unbiased skew over requested axis
# Series.std([axis, skipna, level, ddof, ...])	Return sample standard deviation over requested axis.
# Series.sum([axis, skipna, level, ...])	Return the sum of the values for the requested axis
# Series.var([axis, skipna, level, ddof, ...])	Return unbiased variance over requested axis.
# Series.unique()	Return unique values in the object.
# Series.nunique([dropna])	Return number of unique elements in the object.
# Series.is_unique	Return boolean if values in the object are unique
# Series.is_monotonic	Return boolean if values in the object are
# Series.is_monotonic_increasing	Return boolean if values in the object are
# Series.is_monotonic_decreasing	Return boolean if values in the object are
# Series.value_counts([normalize, sort, ...])	Returns object containing counts of unique values.


# =====================================================================================================
# =====================================================================================================
class GroupBy(GroupByOps):
    """
    Parameters
    ----------
    dataset: Dataset
        The dataset object

    keys: list. List of column names to groupby

    filter: None. Boolean mask array applied as filter before grouping

    return_all: bool.  Default to False. When set to True will return all
          the dataset columns for every operation.

    hint_size: int.  Hint size for the hash (optional)

    sort_display: bool. Default to True.  Indicates

    lex: bool
        Defaults to False. When True uses a lexsort to find the groups (otherwise uses a hash).

    totals: bool
    """

    DebugMode = False
    TestCatGb = True

    def __init__(
        self,
        dataset,
        keys: list = None,
        filter=None,
        ordered=None,
        sort_display=None,
        return_all=False,
        hint_size=0,
        lex: bool = None,
        rec: bool = False,
        totals=False,
        copy=False,
        cutoffs=None,
        verbose=False,
        **kwargs,
    ):

        # upon creation create a unique id to track sorting
        self._uniqueid = GetTSC()

        # short circuit in copy mode
        if copy:
            return

        # copy over sort list
        self._col_sortlist = dataset._col_sortlist

        grouping_dict = None

        if not isinstance(keys, list):
            # if just one str passed in, wrap it in a list
            if isinstance(keys, str):
                keys = [keys]
            else:
                if isinstance(keys, np.ndarray):
                    if len(keys) != dataset.shape[0]:
                        raise TypeError(
                            f'The argument "keys" passed in GroupBy must be the same length as the dataset {len(keys)} vs {dataset.shape[0]}.'
                        )
                    else:
                        name = "col_0"
                        try:
                            name = keys.get_name()
                        except:
                            pass
                        # make into a dict
                        keys = {name: keys}
                        grouping_dict = keys
                else:
                    raise TypeError(
                        f'The argument "keys" passed in GroupBy init must be a list or str.  It must contain column names.'
                    )

        self._sortedbykey = False
        self._isortrows = None
        # self._gbkeys = None
        self._gb_keychain = None

        self._return_all = return_all
        self._filter = filter
        self._pcutoffs = cutoffs
        self._totals = totals

        if grouping_dict is None:
            grouping_dict = {k: dataset[k] for k in keys}

        if ordered is not None and sort_display is not None:
            raise TypeError(f"The kwarg ordered and sort_display cannot be used in combination")

        if ordered is not None:
            sort_display = ordered

        if sort_display is None:
            # automatically sort display if nothing passed
            if ordered is None and lex is None:
                sort_display = True
                try:
                    # if the first key has a grouping object and it does not want the display sorted and is itself not sorted,
                    # then we do not sort the display
                    # this happens after a 'cut' where we want to preserve label order
                    for v in grouping_dict.values():
                        if not v.grouping.isdisplaysorted and not v.grouping.isordered:
                            sort_display = False
                        break
                except:
                    pass

        self.grouping = TypeRegister.Grouping(
            grouping_dict,
            hint_size=hint_size,
            filter=filter,
            sort_display=sort_display,
            lex=lex,
            rec=rec,
            cutoffs=cutoffs,
            verbose=verbose,
        )

        # always use the names of the keys in the grouping dict
        # grouping dict should only live in grouping
        gvals = list(self.grouping._anydict.values())
        if len(keys) == len(gvals):
            self._grouping_dict = {k: gvals[i] for i, k in enumerate(keys)}
        # grouping dict may be larger than keys (multikey categorical splits into multiple arrays)
        else:
            self._grouping_dict = self.grouping._anydict

        # todo: remember change count on dataset
        self._dataset = dataset

        self._sort_display = sort_display

    # ---------------------------------------------------------------
    def copy(self, deep=True):
        """
        Called from getitem when user follows gb with []
        """
        newgb = type(self)(None, copy=True)
        newgb._col_sortlist = self._col_sortlist
        newgb._sortedbykey = self._sortedbykey
        try:
            newgb._gb_keychain = self._gb_keychain.copy(deep=deep)
        except:
            newgb._gb_keychain = None

        newgb._grouping_dict = self._grouping_dict.copy()
        newgb.grouping = self.grouping.copy(deep=deep)
        if deep:
            newgb._dataset = self._dataset.copy()
            if self._filter is None:
                newgb._filter = self._filter
            else:
                newgb._filter = self._filter.copy()

            # newgb.grouping = self.grouping.copy(grouping_dict=newgb._grouping_dict)

        else:
            newgb._dataset = self._dataset
            newgb._filter = self._filter

            # reference vs copy
            # newgb.grouping = self.grouping

        newgb._return_all = self._return_all
        newgb._pcutoffs = self._pcutoffs

        newgb._sort_display = self._sort_display
        # newgb._gbkeys = self._gbkeys
        newgb._isortrows = self._isortrows
        newgb._totals = self._totals

        return newgb

    # ---------------------------------------------------------------
    @property
    def gbkeys(self):
        """
        dictionary of numpy arrays binned from
        """
        return self.gb_keychain.gbkeys

    # ---------------------------------------------------------------
    @property
    def isortrows(self):
        """
        sorted index or None
        """
        return self.gb_keychain.isortrows

    # ---------------------------------------------------------------
    @property
    def ifirstkey(self):
        return self.grouping.ifirstkey

    # ---------------------------------------------------------------
    @property
    def ilastkey(self):
        return self.grouping.ilastkey

    # ---------------------------------------------------------------
    @property
    def gb_keychain(self):
        if self._gb_keychain is None:
            # this will change when groupby keys are handled only by the Grouping object
            # self._gb_keychain = GroupByKeys(self.grouping._anydict, self.ifirstkey, sort_display=self._sort_display, prebinned=self.grouping.iscategorical)
            # always send in the uniques
            self._gb_keychain = GroupByKeys(
                self.grouping.uniquedict, None, sort_display=self._sort_display, prebinned=True
            )
        return self._gb_keychain

    # ---------------------------------------------------------------
    def __iter__(self):
        """
        Generates tuples of key, value pairs.
        Keys are key values for single key, or tuples of key values for multikey.
        Values are datasets containing all rows from data in group for that key.
        """
        return self._iter_internal_contiguous()

    # ---------------------------------------------------------------
    def _pop_gb_data(self, calledfrom, userfunc, *args, **kwargs):
        """
        GroupBy holds on to its dataset. There may be no additional data provided.
        """
        kwargs.setdefault("dataset", self._dataset)
        return super()._pop_gb_data(calledfrom, userfunc, *args, **kwargs)

    # ---------------------------------------------------------------
    def add_totals(self, gb_ds):
        total = {}
        labels = gb_ds.label_get_names()
        for colname, arr in gb_ds.items():
            if colname not in labels:
                if arr.iscomputable():
                    total[colname] = arr.sum()
                else:
                    # leave blank
                    total[colname] = ""

        gb_ds.footer_set_values("Total", total)
        return gb_ds

    # ---------------------------------------------------------------
    # OVERRIDEN from groupbyops
    def _calculate_all(self, funcNum, *args, func_param=0, **kwargs):
        """
        Generate a GroupByKeys object if necessary and ask for the result of a calculation from the grouping object.
        Returns: a grouped by dataset with the result from the calculation
        """
        keychain = self.gb_keychain
        kwargs.setdefault("dataset", self._dataset)
        return_all = kwargs.pop("return_all", self._return_all)

        origdict, user_args, tups = self._prepare_gb_data("GroupBy", funcNum, *args, **kwargs)
        gb_ds = self.grouping._calculate_all(
            origdict,
            funcNum,
            func_param=func_param,
            return_all=return_all,
            keychain=keychain,
            user_args=user_args,
            tups=tups,
            **kwargs,
        )

        # check for accum1 like functionality
        if self._totals:
            return self.add_totals(gb_ds)
        else:
            return self._possibly_transform(gb_ds, label_keys=keychain.keys(), **kwargs)

    # ---------------------------------------------------------------
    def count(self, **kwargs):
        """Compute count of group"""
        return self.grouping.count(keychain=self.gb_keychain, **kwargs)

    ## ------------------------------------------------------------
    def __getattr__(self, name):
        """
        __getattr__ is hit when '.' is used to trim a single column.

        Examples
        --------
        >>> ds = Dataset({'col_'+str(i): np.random.rand(5) for i in range(5)})
        >>> ds.keycol = FA(['a','a','b','c','a'])
        >>> ds.gb('keycol').col_4.mean()
        *keycol   col_4
        -------   -----
        a          0.73
        b          0.03
        c          0.76

        """
        # check if the name exists in the dataset
        if name[0] == "_" or name not in self._dataset:
            return super().__getattribute__(name)

        return self._getitem(name)

    # ---------------------------------------------------------------
    def as_categorical(self):
        """
        Returns a categorical using the same binning information as the GroupBy object (no addtl. hash required).
        New categorical will not share a grouping object with this groupby object, but will share a reference to the iKey.
        Categorical operation results will be sorted or unsorted depending on if 'gb' or 'gbu' called this.
        """

        idx = self.grouping.ikey
        gbkeys = self.gbkeys

        return TypeRegister.Categorical(idx, _from_categorical=gbkeys, sort_gb=self._sort_display)

    # ---------------------------------------------------------------
    def _grouping_data_as_dict(self, ds):
        return self._dataset.as_ordered_dictionary()

    # -------------------------------------------------------
    def pad(self, limit=0, fill_val=None, inplace=False):
        """
        Forward fill the values

        Parameters
        ----------
        limit : integer, optional
            limit of how many values to fill

        See Also
        --------
        fill_forward
        fill_backward
        fill_invalid
        """
        return self.fill_forward(fill_val=fill_val, limit=limit, inplace=inplace)

    # -------------------------------------------------------
    def fill_forward(self, limit=0, fill_val=None, inplace=False):
        """
        Replace NaN and invalid array values by propagating the last encountered valid
        group value forward.

        Parameters
        ----------
        limit : int, default 0
            The maximium number of consecutive NaN or invalid values to fill. If there
            is a gap with more than this number of consecutive NaN or invalid values,
            the gap will be only partially filled. If no `limit` is specified, all
            consecutive NaN and invalid values are replaced.
        fill_val : scalar, default None
            The value to use where there is no valid group value to propagate forward.
            If `fill_val` is not specified, NaN and invalid values aren't replaced where
            there is no valid group value to propagate forward.
        inplace : bool, default False
            If False, return a copy of the array. If True, modify original data. This
            will modify any other views on this object. This fails if the array is
            locked.

        Returns
        -------
        `Dataset`
            The returned `Dataset` contains the input `Dataset` object's numerical
            columns.

        See Also
        --------
        GroupBy.fill_backward :
            Replace NaN and invalid array values with the next valid group value.
        Categorical.fill_forward :
            Replace NaN and invalid array values with the last valid group value.
        riptable.fill_forward : Replace NaN and invalid values with the last valid
            value.
        Dataset.fillna : Replace NaN and invalid values with a specified value or
            nearby data.
        FastArray.fillna : Replace NaN and invalid values with a specified value or
            nearby data.
        FastArray.replacena : Replace NaN and invalid values with a specified value.

        Examples
        --------
        >>> ds = rt.Dataset({'Key_col' : ['A', 'B', 'A', 'B', 'A', 'B'],
        ...                  'Vals' : [0, 1, 2, 3, rt.nan, rt.nan]})
        >>> ds.gb('Key_col').fill_forward()
        #   Vals
        -   ----
        0   0.00
        1   1.00
        2   2.00
        3   3.00
        4   2.00
        5   3.00

        Use a `fill_val` to replace values where there's no valid group value to
        propagate forward:

        >>> ds.Vals = rt.FastArray([rt.nan, rt.nan, 2, 3, 4, 5])
        >>> ds.gb('Key_col').fill_forward(fill_val = 0)
        #   Vals
        -   ----
        0   0.00
        1   0.00
        2   2.00
        3   3.00
        4   4.00
        5   5.00


        Replace only the first NaN or invalid value in any consecutive series of NaN or
        invalid values in a group:

        >>> ds.Vals = rt.FastArray([0, 1, rt.nan, rt.nan, rt.nan, rt.nan])
        >>> ds.gb('Key_col').fill_forward(limit = 1)
        #   Vals
        -   ----
        0   0.00
        1   1.00
        2   0.00
        3   1.00
        4    nan
        5    nan
        """
        return super().nb_fill_forward(fill_val=fill_val, limit=limit, inplace=inplace)

    # -------------------------------------------------------
    def backfill(self, limit=0, fill_val=None, inplace=False):
        """
        Backward fill the values

        Parameters
        ----------
        limit : integer, optional
            limit of how many values to fill

        See Also
        --------
        fill_forward
        fill_backward
        fill_invalid
        """
        # stub for pandas
        return self.fill_backward(fill_val=fill_val, limit=limit, inplace=inplace)

    # -------------------------------------------------------
    def fill_backward(self, limit=0, fill_val=None, inplace=False):
        """
        Replace NaN and invalid array values by propagating the next encountered valid
        group value backward.

        Parameters
        ----------
        limit : int, default 0
            The maximium number of consecutive NaN or invalid values to fill. If there
            is a gap with more than this number of consecutive NaN or invalid values,
            the gap will be only partially filled. If no `limit` is specified, all
            consecutive NaN and invalid values are replaced.
        fill_val : scalar, default None
            The value to use where there is no valid group value to propagate backward.
            If `fill_val` is not specified, NaN and invalid values aren't replaced where
            there is no valid group value to propagate backward.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        `Dataset`
            The returned `Dataset` contains the input `Dataset` object's numerical
            columns.

        See Also
        --------
        GroupBy.fill_forward :
            Replace NaN and invalid array values with the last valid group value.
        Categorical.fill_backward :
            Replace NaN and invalid array values with the next valid group value.
        riptable.fill_backward : Replace NaN and invalid values with the next valid
            value.
        Dataset.fillna : Replace NaN and invalid values with a specified value or
            nearby data.
        FastArray.fillna : Replace NaN and invalid values with a specified value or
            nearby data.
        FastArray.replacena : Replace NaN and invalid values with a specified value.

        Examples
        --------
        >>> ds = rt.Dataset({'Key_col' : ['A', 'B', 'A', 'B', 'A', 'B'],
        ...             'Vals' : [rt.nan, rt.nan, 2, 3, 4, 5]})
        >>> ds.gb('Key_col').fill_backward()
        #   Vals
        -   ----
        0   2.00
        1   3.00
        2   2.00
        3   3.00
        4   4.00
        5   5.00

        Use a `fill_val` to replace values where there's no valid group value to
        propagate backward:

        >>> ds.Vals = rt.FastArray([0, 1, 2, 3, rt.nan, rt.nan])
        >>> ds.gb('Key_col').fill_backward(fill_val = 0)
        #   Vals
        -   ----
        0   0.00
        1   1.00
        2   2.00
        3   3.00
        4   0.00
        5   0.00

        Replace only the first NaN or invalid value in any consecutive series of NaN or
        invalid values in a group:

        >>> ds.Vals = rt.FastArray([rt.nan, rt.nan, rt.nan, rt.nan, 4, 5])
        >>> ds.gb('Key_col').fill_backward(limit = 1)
        #   Vals
        -   ----
        0    nan
        1    nan
        2   4.00
        3   5.00
        4   4.00
        5   5.00
        """
        return super().nb_fill_backward(fill_val=fill_val, limit=limit, inplace=inplace)

    # -------------------------------------------------------
    def expanding(self, **kwargs):
        raise NotImplementedError("riptable uses the Grouping object (see .group)ck")

    # -------------------------------------------------------
    def stack(self, **kwargs):
        raise NotImplementedError("riptable uses the Accum2 class for stack/unstack")

    # -------------------------------------------------------
    def unstack(self, **kwargs):
        raise NotImplementedError("riptable uses the Accum2 class for stack/unstack")

    # -------------------------------------------------------
    @property
    def transform(self):
        """
        The property transform sets a flag so that the next reduce function called after transform,
        will repopulate the original array with the reduced value.

        Example
        -------
        >>> ds.groupby(['side', 'venue']).transform.sum()
        """
        warnings.warn("Deprecation warning: Use kwarg transform=True instead of transform.")
        self._transform = True
        return self

    # -------------------------------------------------------
    def get_group(self, category, **kwargs):
        """
        The name of the group to get as a Dataset.

        Parameters
        ----------
        category: string or tuple
            A value from the column used to construct the GroupBy, or if
            multiple columns were used, a tuple of the multiple columns.

        Returns
        -------
        Dataset

        Example
        -------
        >>> ds.groupby('symbol').get_group('AAPL')
        """
        # categorical has method to get the bin
        cat = self.as_categorical()

        bin = cat.from_category(category)
        return self._dataset[self.grouping.ikey == bin, :]

    # -------------------------------------------------------
    def _build_string(self):
        newString = f"GroupBy Keys {list(self._grouping_dict.keys())} @ [{self._dataset._ncols} x {self.grouping.unique_count}]\n"
        newString += f"ikey:{not isinstance(self.grouping.ikey, type(None))}  iFirstKey:{not isinstance(self.grouping.iFirstKey, type(None))}  iNextKey:{not isinstance(self.grouping.iNextKey, type(None))}  nCountGroup:{not isinstance(self.grouping.nCountGroup, type(None))} "
        # newString += f" _isortrows:{not isinstance(self._isortrows, type(None))}  _gbkeys:{not isinstance(self._gbkeys, type(None))}  _filter:{not isinstance(self._filter, type(None))}  _return_all:{self._return_all}\n\n"

        newString += f"_filter:{not isinstance(self._filter, type(None))}  _return_all:{self._return_all}\n\n"

        # temporarily build a dataset
        dset = self.count()

        # dump of count
        resultString = dset.__str__()

        return newString + resultString

    # -------------------------------------------------------
    def __repr__(self):
        return self._build_string()

    # -------------------------------------------------------
    def __str__(self):
        return self._build_string()

    # -------------------------------------------------------
    def _getitem(self, fld):
        """
        Called by __getitem__ and __getattr__.
        Uses the field to index into the stored dataset.
        Often used to limit the data the groupby operation is being performed on.
        Returns a shallow copy of the groupby object.

        This routine gets hit during the following common code pattern:

        >>> ds = Dataset({'col_'+str(i): np.random.rand(5) for i in range(5)})
        >>> ds.keycol = FA(['a','a','b','c','a'])
        >>> ds.gb('keycol')[['col_1', 'col_2']].sum()
        *keycol   col_1   col_2
        -------   -----   -----
        a          1.92    0.89
        b          0.70    0.46
        c          0.07    0.42

        >>> ds.gb('keycol').col_4.mean()
        *keycol   col_4
        -------   -----
        a          0.73
        b          0.03
        c          0.76

        """
        # return new groupby
        newgb = self.copy(deep=False)

        # reduce the datasets
        if isinstance(fld, (str, bytes)):
            # need a dataset back, not a fastarray
            index = [fld]
        elif isinstance(fld, list):
            index = fld.copy()
        # turn to list of multiple strings
        elif isinstance(fld, tuple):
            index = list(fld)
        else:
            raise TypeError("Dont know how to make groupby from", fld)

        # we must add back in the grouping keys or we cannot group
        for item in self._grouping_dict.keys():
            if item not in index:
                index.append(item)

        newgb._dataset = newgb._dataset[index]
        return newgb

    # -------------------------------------------------------
    def __getitem__(self, fld):
        return self._getitem(fld)


# keep this as the last line
from .rt_enum import TypeRegister

TypeRegister.GroupBy = GroupBy



__all__ = ["Accum2"]

import warnings

import numpy as np

from .rt_categorical import Categorical
from .rt_enum import (
    FILTERED_LONG_NAME,
    GB_FUNC_COUNT,
    GB_FUNCTIONS,
    INVALID_DICT,
    INVALID_SHORT_NAME,
    DisplayJustification,
    DisplayLength,
    TypeId,
    TypeRegister,
)
from .rt_fastarray import FastArray
from .rt_groupbyops import GroupByOps
from .rt_grouping import combine2groups
from .rt_numpy import (
    _groupbycalculateall,
    _groupbycalculateallpack,
    bool_to_fancy,
    empty,
    full,
    groupbypack,
    ismember,
    ones,
    where,
    zeros,
)
from .Utils.rt_display_properties import ItemFormat
from .Utils.rt_metadata import MetaData


class Accum2(GroupByOps, FastArray):
    """
    The Accum2 object is very similar to a GroupBy object that has been initialized with a multikey Categorical.

    The Accum2 object is very similar to a GroupBy object that has been initialized with a multikey Categorical.
    Because it also inherits from GroupByOps, all calculations will be sent to _calculate_all in a Grouping object.
    Accum2 generates a single array of data, and splits it into multiple columns - one for each x-axis bin.
    There is always an invalid bin, but it is omitted by default when the single array is split into columns.
    Datasets resulting from an Accum2 groupby calculation will be displayed with a footer row of column totals,
    and an additional vertical column of row totals.

    In addition to inheriting from GroupByOps, Accum2 also inherits from FastArray. This way, it can exist as a
    column in a Dataset. Its cell data will appear as a tuple of values from its X and Y axis.

    Parameters
    ----------
    cat_rows: Categorical
        Categorical for the rows axis, or an array which will be converted to a Categorical.

    cat_cols: Categorical
        Categorical for the column axis, or an array which will be converted to a Categorical.

    Keywords
    --------
    invalid: defaults to False. Set to True to show filtered columns
    ordered: defaults to None. See Categorical
    sort_gb: defaults to False. See Categorical
    ylabel: defaults to None. Set to a string to override the name of the left column
    totals: defaults to True.
    There is no sort_display option

    Returns
    -------
    Accum2 object which can be used to perform calculations
    Accum2 subclasses from FastArray and can be added to a dataset

    Accum2.operation is then supported.  Accum2(catx, caty).min(array1)
    See: groupbyops

    Examples
    --------
    >>> int_fa = FastArray([1,2,3,4]*4)
    >>> str_fa = FastArray(['a','b','c','d','b','c','d','a','c','d','b','a','d','a','b','c'])
    >>> data_col = np.random.rand(16)*10
    >>> data_col
    array([6.7337479 , 1.69561884, 8.20657899, 6.12821287, 3.95380641,
            1.06706672, 9.51679965, 3.57184704, 7.86268264, 9.0136061 ,
            2.12355667, 3.64954958, 8.40952542, 0.06431684, 9.52872172,
            3.94938333])   #random

    >>> c_x = Categorical(str_fa)
    >>> c_y = Categorical(int_fa)
    >>> ac = Accum2(c_x, c_y)
    >>> ac
    Accum2 Keys
     X:[b'a' b'b' b'c' b'd']
     Y:{'key_0': FastArray([1, 2, 3, 4])}
     Bins:25   Rows:16
    <BLANKLINE>
    *YLabel   a   b   c   d   Total
    -------   -   -   -   -   -----
          1   1   1   1   1       4
          2   1   1   1   1       4
          3   0   2   1   1       4
          4   2   0   1   1       4
    -------   -   -   -   -   -----
      Total   4   4   4   4      16

    >>> ac.sum(data_col)
    *YLabel       a       b       c       d   Total
    -------   -----   -----   -----   -----   -----
          1    6.73    3.95    7.86    8.41   26.96
          2    0.06    1.70    1.07    9.01   11.84
          3    0.00   11.65    8.21    9.52   29.38
          4    7.22    0.00    3.95    6.13   17.30
    -------   -----   -----   -----   -----   -----
      Total   14.02   17.30   21.09   33.07   85.48
    """

    DebugMode: bool = False
    # max value set for x-axis labels (multikey only). performance will suffer if large array of tuple strings is generated.
    ACCUM_X_MAX: int = 10_000

    def __new__(
        cls, cat_rows, cat_cols, filter=None, showfilter=False, ordered=None, sort_gb=False, totals=True, ylabel=None
    ):
        # sort_display/sort_gb is not allowed so that the imatrix is always correct
        try:
            if not isinstance(cat_rows, Categorical):
                cat_rows = Categorical(cat_rows, ordered=ordered, sort_gb=sort_gb)
        except:
            pass

        if not isinstance(cat_rows, Categorical):
            raise TypeError(
                f"accum2: Argument 1 must be a categorical or an array that can be made into a categorical not type {type(cat_rows)!r}"
            )

        try:
            if not isinstance(cat_cols, Categorical):
                cat_cols = Categorical(cat_cols, ordered=ordered, sort_gb=sort_gb)
        except:
            pass

        if not isinstance(cat_cols, Categorical):
            raise TypeError(
                f"accum2: Argument 2 must be a categorical or an array that can be made into a categorical not type {type(cat_cols)!r}"
            )

        # enum or dict type categoricals are not 0 or 1 based, so we convert them
        if cat_rows.isenum:
            cat_rows = cat_rows.as_singlekey(ordered=ordered)

        if cat_cols.isenum:
            cat_cols = cat_cols.as_singlekey(ordered=ordered)

        # test uniqueness of cols categorical - large multikey will cause performance errors
        col_keys = cat_cols.gb_keychain
        if col_keys.multikey:
            if col_keys.unique_count > cls.ACCUM_X_MAX:
                raise ValueError(
                    f"Multikey categorical's groupby keys are too large for column-axis. Use a unique amount smaller than {cls.ACCUM_X_MAX} or use in the row-axis"
                )

        # this will group the two categoricals
        # generate iKey,unique_count and grouping object

        # call CPP algo to merge two bins into one
        grouping = combine2groups(cat_rows.grouping, cat_cols.grouping, filter=filter)
        instance = grouping.ikey.view(cls)
        instance.grouping = grouping
        instance._cat_cols = cat_cols
        instance._cat_rows = cat_rows
        instance._return_all = False
        instance._showfilter = showfilter
        instance._totals = totals
        instance._dataset = None
        instance._gb_keychain = None
        instance._ylabel = ylabel
        # filter already applied but groupbyops looks for it by default
        instance._filter = None
        # _myfilter remembers the filter, since _filter is always set to None to work with groupbyops
        instance._myfilter = filter

        return instance

    # ---------------------------------------------------------------
    def __init__(
        cls, cat_rows, cat_cols, filter=None, showfilter=False, ordered=None, sort_gb=False, totals=True, ylabel=None
    ):
        pass

    # ------------------------------------------------------------
    def __len__(self):
        return super().__len__()

    # ------------------------------------------------------------
    def __del__(self):
        """
        Called when a Categorical is deleted.
        """
        # python has trouble deleting objects with circular references
        del self.grouping
        self.grouping = None
        del self._cat_cols
        del self._cat_rows
        if self._dataset is not None:
            del self._dataset
        if self._gb_keychain is not None:
            del self._gb_keychain
        if self._ylabel is not None:
            del self._ylabel

    # ------------------------------------------------------------
    @property
    def size(self):
        return self.__len__()

    # ---------------------------------------------------------------
    @property
    def gbkeys(self):
        return self._cat_rows.grouping_dict

    # ---------------------------------------------------------------
    @property
    def isortrows(self):
        return self._cat_rows.isortrows

    # ---------------------------------------------------------------
    @property
    def gb_keychain(self):
        """
        Request a GroupByKeys from the y-axis categorical.

        This provides unique keys, a possible sorted index, and the ability to add a filtered bin to the final table from groupby calculations.
        """
        if self._gb_keychain is None:
            self._gb_keychain = self._cat_rows.gb_keychain
        return self._gb_keychain

    # ------------------------------------------------------------
    def display_query_properties(self):
        """
        Take over display query properties from parent class FastArray.

        When displayed in a Dataset, Accum2 data will be displayed as a tuple composite of its categorical (x,y) bin values.
        """
        item_format = ItemFormat(
            length=DisplayLength.Long, justification=DisplayJustification.Left, can_have_spaces=True, decoration=None
        )
        convert_func = self.display_convert_func
        return item_format, convert_func

    # ------------------------------------------------------------
    def display_convert_func(self, index, itemformat: ItemFormat):
        # TODO: apply ItemFormat options that were passed in
        item = str(self._internal_getitem(index))
        return str(item).replace("'", "")

    # ------------------------------------------------------------
    def _internal_getitem(self, matrix_index):
        xidx = matrix_index // (self._cat_rows.unique_count + 1)
        yidx = matrix_index % (self._cat_rows.unique_count + 1)

        if xidx == 0:
            xcat = INVALID_SHORT_NAME
        else:
            xidx -= 1
            xcat = self._cat_cols.gb_keychain.get_bin(xidx)

        if yidx == 0:
            ycat = INVALID_SHORT_NAME
        else:
            yidx -= 1
            ycat = self._cat_rows.gb_keychain.get_bin(yidx)

        if isinstance(xcat, bytes):
            xcat = bytes.decode(xcat)
        if isinstance(ycat, bytes):
            ycat = bytes.decode(ycat)
        result = (xcat, ycat)
        return result

    # ------------------------------------------------------------
    def __getitem__(self, fld):
        """
        Bracket indexing for Accum2.
        """
        if Accum2.DebugMode:
            print("***get item from:", fld)

        result = None
        # cat[int]
        # returns single category
        if isinstance(fld, (int, np.integer)):
            if Accum2.DebugMode:
                print("***self.view(FastArray)\n", self.view(FastArray))
            matrix_index = self._np[fld]
            return self._internal_getitem(matrix_index)

        else:
            return super(Accum2, self).__getitem__(fld)

    # ---------------------------------------------------------------
    def _get_gbkeyname(self):
        if self._ylabel is not None:
            return self._ylabel

        # for single key
        gbkeyname = self._cat_rows.get_name()
        if gbkeyname is None:
            gbkeyname = "YLabel"
        return gbkeyname

    # ---------------------------------------------------------------
    def _get_gbkeys(self, showfilter=False):
        if self.gb_keychain.singlekey:
            gbkeyname = self._get_gbkeyname()
            if showfilter:
                keys = self.gb_keychain.gbkeys_filtered
            else:
                keys = self.gb_keychain.gbkeys

            keycol = list(keys.values())[0]
            if self._cat_rows.category_mode in TypeRegister.Categories.dict_modes:
                keycol = Categorical(keycol, _from_categorical=self._cat_rows._categories_wrap)

            return {gbkeyname: keycol}
        # return {gbkeyname:self._cat_rows._categories}
        else:
            if showfilter:
                return self.gb_keychain.gbkeys_filtered
            else:
                return self.gb_keychain.gbkeys

    # ---------------------------------------------------------------
    def _make_imatrix(self, input_arr, col_keys, row_keys, showfilter=False):
        """
        Return a Fortran-ordered 2d matrix. ::

            if showfilter is False, the first column is removed
                shape is (row_keys.unique_count+1, col_keys.unique_count)
            else if showfilter is True
                shape is (row_keys.unique_count +1, col_keys.unique_count+1)
        """
        if showfilter:
            tempi = input_arr[0:-1]
            imatrix = tempi.reshape((row_keys.unique_count + 1, col_keys.unique_count + 1), order="F")
        else:
            # skip over first column (the filtered column)
            tempi = input_arr[row_keys.unique_count + 1 : -1]
            imatrix = tempi.reshape((row_keys.unique_count + 1, col_keys.unique_count), order="F")
        return imatrix

    # ---------------------------------------------------------------
    def make_dataset(self, arr, showfilter=False):
        """

        Parameters
        ----------
        arr: input array of data

        Returns
        -------
        ds
        col_keys
        row_keys
        """
        # default to no showfilter
        showfilter_base = 1

        # check if we need to show showfilter columns
        if showfilter:
            showfilter_base = 0

        # possibly attach filter bin to gbkey columns
        gbkeys = self._get_gbkeys(showfilter=showfilter)

        # put the grouping keys as the first columns
        newds = TypeRegister.Dataset(gbkeys)

        col_keys = self._cat_cols.gb_keychain
        row_keys = self._cat_rows.gb_keychain

        # x-axis headers need to be a single list
        if col_keys.singlekey:
            # xcategories = self._cat_cols._categories
            xcategories = list(col_keys.gbkeys.values())[0]
        else:
            # generate tuple strings for multikey
            if showfilter:
                xcategories = col_keys.multikey_labels_filtered
            else:
                xcategories = col_keys.multikey_labels

        offsety = row_keys.unique_count + 1

        if Accum2.DebugMode:
            print("**stack len", len(arr), offsety, "  showfilterbase:", showfilter_base)

        # add showfilter column first if we have to
        if showfilter:
            newds[FILTERED_LONG_NAME] = arr[0:offsety]

        # skip showfilter row for loop below (already added or not added above)
        offset = offsety

        # fix bug for enums, need to reattach code mapping for correct string
        xmode = self._cat_cols.category_mode
        if xmode in TypeRegister.Categories.dict_modes:
            xcategories = TypeRegister.Categorical(xcategories, _from_categorical=self._cat_cols._categories_wrap)

        # cut main array into multiple columns
        for i in range(col_keys.unique_count):
            new_colname = xcategories[i]

            if isinstance(new_colname, bytes):
                new_colname = new_colname.decode()

            if isinstance(new_colname, str):
                if len(new_colname) == 0:
                    # make up a column name
                    new_colname = INVALID_SHORT_NAME + str(i)
            else:
                new_colname = str(new_colname)

            start = showfilter_base + offset
            stop = offset + offsety
            offset += offsety

            # possibly skip over filter
            arridx = slice(start, stop)
            newds[new_colname] = arr[arridx]

        return {"ds": newds, "col_keys": col_keys, "row_keys": row_keys, "gbkeys": gbkeys}

    # ---------------------------------------------------------------
    @classmethod
    def _apply_2d_operation(self, func, imatrix, showfilter=True, filter_rows=None, filter_cols=None):
        """
        Called from routines like sum or min where we can make one pass

        If there are badrows, then filter_rows is set to the row indexes that are bad
        If there are badcols, then filter_cols is set to the col indexes that are bad
        filter_rows is a fancy index or none
        """

        if callable(func):

            row_count, col_count = imatrix.shape

            # horizontal add
            # print("im0", imatrix.nansum())
            totalsY = func(imatrix, axis=1)  # [showfilter_base:]

            # vertical func operation
            totalsX = empty(col_count, dtype=totalsY.dtype)

            # possibly remove filtered top row
            if not showfilter:
                totalsY = totalsY[1:]

            # consider #imatrix.nansum(axis=0, out=totalsX)
            for i in range(col_count):
                arrslice = imatrix[:, i]

                # possibly skip over first value
                if not showfilter:
                    arrslice = arrslice[1:]

                totalsX[i] = func(arrslice)

            return totalsX, totalsY

        # function was not callable
        return None, None

    # ---------------------------------------------------------------
    @classmethod
    def _accum1_pass(cls, cat, origarr, funcNum, showfilter=False, filter=None, func_param=0, **kwargs):
        """
        internal call to calculate the Y or X summary axis
        the filter muse be passed correctly
        returns array with result of operation, size of array is number of uniques
        """

        basebin = 1
        if showfilter:
            basebin = 0

        if callable(funcNum):
            # from apply_reduce
            # funcList = [GB_FUNCTIONS.GB_SUM]
            # accum_tuple = _groupbycalculateall([origarr], ikey, numkeys, funcList, binLowList, binHighList, func_param)

            # need a new option here, which is that we want to allocate for a filter
            # but we might not use it
            # ALSO dont want back a dataaset
            accum_tuple = cat.apply_reduce(
                funcNum, origarr, showfilter=showfilter, filter=filter, nokeys=True, **kwargs
            )

            # the showfilter is handled automatically
            return accum_tuple[0]

        else:
            ikey = cat.grouping.ikey

            # if zero base, we need 1 base for these calculations
            if cat.grouping.base_index == 0:
                ikey = ikey + 1

            # Optimization: combine_filter was previously called
            if filter is not None:
                # N.B. We are going to change ikey, make a copy instead of changing the input. The input
                #      data will be used again when user calls method on the Accum2 object again.
                # zero out anything not in the filter
                ikey = where(filter, ikey, 0)

            numkeys = cat.unique_count

            funcList = [funcNum]

            binLowList = [basebin]
            binHighList = [numkeys + 1]

            if funcNum >= GB_FUNCTIONS.GB_SUM and funcNum < GB_FUNCTIONS.GB_FIRST:
                accum_tuple = _groupbycalculateall(
                    [origarr], ikey, numkeys, funcList, binLowList, binHighList, func_param
                )

            elif funcNum >= GB_FUNCTIONS.GB_FIRST and funcNum < GB_FUNCTIONS.GB_CUMSUM:
                # TODO break out as function
                packing = groupbypack(ikey, None, numkeys + 1)
                iGroup = packing["iGroup"]
                iFirstGroup = packing["iFirstGroup"]
                nCountGroup = packing["nCountGroup"]
                accum_tuple = _groupbycalculateallpack(
                    [origarr],
                    ikey,
                    iGroup,
                    iFirstGroup,
                    nCountGroup,
                    numkeys,
                    funcList,
                    binLowList,
                    binHighList,
                    func_param,
                )

        # whether or not they want to see the filter
        if basebin != 0:
            return accum_tuple[0][basebin:]
        else:
            return accum_tuple[0]

    # ---------------------------------------------------------------
    @classmethod
    def _add_totals(cls, cat_rows, newds, name, totalsX, totalsY, totalOfTotals):
        """
        Adds a summary column on the right (totalsY)
        Adds a footer on the bottom (totalsX)
        """
        if totalsY is not None:
            if newds.shape[0] != len(totalsY):
                # this path is from custom apply_reduce
                emptyarr = empty((newds.shape[0],), dtype=totalsY.dtype)
                emptyarr[0 : len(totalsY)] = totalsY
                emptyarr[-1] = totalOfTotals
                newds[name] = emptyarr

            else:
                # add the Total column to the dataset
                newds[name] = totalsY

            # add to the right summary
            newds.summary_set_names([name])

            # tell display that this dataset has a footer
            # have to skip over the colkeys
            keycount = len(cat_rows.gb_keychain.gbkeys)

            # totalsX runs in the horizontal direction on the bottom
            # for each column name in the dictionary, give a value
            footerdict = dict(zip([*newds][keycount:], totalsX))

            # lower right corner sometimes passed alone
            if totalOfTotals is not None:
                footerdict[name] = totalOfTotals

            newds.footer_set_values(name, footerdict)

    # ---------------------------------------------------------------
    @classmethod
    def _calc_badslots(cls, cat, badslots, filter, wantfancy):
        """
        internal routine
        will combine (row or col filter) badslots with common filter

        if there are not badslots, the common filter is returned
        otherwise a new filter is returned
        the filter is negative (badslots locations are false)

        if wantfancy is true, returns fancy index to cols or rows
        otherwise full boolean mask combined with existing filter (if exists)
        """
        if badslots is None:
            if wantfancy:
                return None
            # user did not pass in any, stick with current filter
            return filter

        badslots = np.atleast_1d(badslots)
        dtypenum = badslots.dtype.num

        if wantfancy:
            # find out which columns are to be filtered out
            # are cols an integer or a string?
            if dtypenum > 10:
                _, newfilter = cat.grouping.ismember(badslots, reverse=True)
            else:
                # assume user passed in row or col numbers that are bad
                # such as badrows=[3,4]
                newfilter = badslots
            return newfilter

        # are they passing in a boolean filter?
        if dtypenum == 0:
            # convert bool mask to row numbers and use that mask
            badslots = bool_to_fancy(badslots)

        if dtypenum <= 10:
            # assumes there is not Cat of integers..otherwise ambiguous
            # add 1 because of base_index
            # should we check showfilter?
            badslots = badslots + 1
            if len(badslots) == 1:
                newfilter = cat._fa != badslots[0]
            else:
                newfilter, _ = ismember(cat._fa, badslots)
                # inplace logical not (this is a negative filter)
                np.logical_not(newfilter, out=newfilter)
        else:
            # create filter
            newfilter = cat.isin(badslots)
            # inplace logical not (this is a negative filter)
            np.logical_not(newfilter, out=newfilter)

        if filter is not None:
            # combine both filters using inplace and of filter
            np.logical_and(newfilter, filter, out=newfilter)

        # print('newfilter', len(newfilter), newfilter.sum(), newfilter)
        # return a new filter
        return newfilter

    # ---------------------------------------------------------------
    @classmethod
    def _calc_multipass(
        cls,
        cat_cols,
        cat_rows,
        newds,
        origarr,
        funcNum,
        func,
        imatrix,
        name=None,
        showfilter=False,
        filter=None,
        badrows=None,
        badcols=None,
        badcalc=True,
        **kwargs,
    ):
        """
        For functions that require multiple passes to get the proper result.
        such as mean or median.

        If the grid is 7 x 11: there will be 77 + 11 + 7 + 1 => 96 passes

        Other Parameters
        ----------------
        func: userfunction to call calculate
        name: optional column name (otherwise function name used)
        badrows: optional list of bad row keys, will be combined with filter
        badcols: optional list of bad col keys, will be combined with filter

        badrows/cols is just the keys that are bad (not a boolean filter)
        for example badrows=['AAPL','GOOG']

        Need new algo to take:
            bad bins + ikey + existing boolean filter ==> create a new boolean filter
            walk ikey, see if bin is bad in lookup table, if so set filter to False
            else copy from existing filter value
        """
        if name is None:
            name = str.capitalize(func.__name__)

        # get a negative boolean filter
        newfilterX = cls._calc_badslots(cat_cols, badcols, filter, False)
        newfilterY = cls._calc_badslots(cat_rows, badrows, filter, False)

        newfilter = None
        # first check for any row and col filters
        if badrows is not None or badcols is not None:

            # the common filter is already merged into the row or col filter
            if badrows is not None and badcols is not None:

                # both col and row filter are in use so combine the filters
                newfilter = newfilterX & newfilterY
            else:
                if badrows is not None:
                    newfilter = newfilterY
                else:
                    newfilter = newfilterX
        else:
            newfilter = filter

        # if there is not filter, the value will be None
        if Accum2.DebugMode:
            print("filterrows", newfilterY)
            print("filtercols", newfilterX)
            print("filter    ", newfilter)

        # set to False so that totalsX has invalid where the badcols are
        # set to False so that totalsY has invalid where the badrows are
        # badcalc =True

        if badcalc:
            #  pass in original filter
            totalsX = cls._accum1_pass(cat_cols, origarr, funcNum, showfilter=showfilter, filter=newfilterY, **kwargs)
            totalsY = cls._accum1_pass(cat_rows, origarr, funcNum, showfilter=showfilter, filter=newfilterX, **kwargs)

        else:
            #  pass in combined filter since the filter was handled on class
            totalsX = cls._accum1_pass(cat_cols, origarr, funcNum, showfilter=showfilter, filter=newfilter, **kwargs)
            totalsY = cls._accum1_pass(cat_rows, origarr, funcNum, showfilter=showfilter, filter=newfilter, **kwargs)

        # calculate total of totals
        if func is not None:
            # we can have common filters, row filters, and col filters

            try:
                if newfilter is not None:
                    totalOfTotals = func(origarr[newfilter])
                else:
                    totalOfTotals = func(origarr)
            except:
                if "func_param" not in kwargs:
                    raise ValueError(f"Function {name} requires an extra parameter.")
                func_param = kwargs["func_param"]
                if newfilter is not None:
                    totalOfTotals = func(origarr[newfilter], func_param)
                else:
                    totalOfTotals = func(origarr, func_param)
        else:
            # todo, get invalid
            totalOfTotals = 0

        cls._add_totals(cat_rows, newds, name, totalsX, totalsY, totalOfTotals)

    # ---------------------------------------------------------------
    @classmethod
    def _calc_onepass(
        cls,
        cat_cols,
        cat_rows,
        newds,
        origarr,
        funcNum,
        func,
        imatrix,
        name=None,
        showfilter=False,
        filter=None,
        badrows=None,
        badcols=None,
        badcalc=True,
        **kwargs,
    ):
        """
        For functions such as sum or min that require one pass to get the proper result.

        The first pass calculates all the cells.  Once the cells are calculated,
        an imatrix is made.  Since functions like sum or min can calculate proper values
        for horizontal or vertical operations without making another pass, we use
        the imatrix to calculate the rest.

        The user may also pass in badrows or badcols, or both.
        When badrows is passed, the CELLS for that row are still calculated normally.
        However, the totalOfTotals will not include the badrows or cols.
        """

        # to make sure column names do not conflict with methods we capitalize first letter
        if name is None:
            name = str.capitalize(func.__name__)

        newfilter_cols = cls._calc_badslots(cat_cols, badcols, filter, True)
        newfilter_rows = cls._calc_badslots(cat_rows, badrows, filter, True)

        # newfilter_cols/rows is a fancy index
        if Accum2.DebugMode:
            print("newfilter_cols", newfilter_cols)
            print("newfilter_rows", newfilter_rows)

        # do both horizontal and vertical calculations which are clean
        # TODO optimization -- can just use empty for totalsY and not calculate it if
        #                       newfilter_rows is set  (same for totalsX)
        totalsX, totalsY = cls._apply_2d_operation(
            func, imatrix, showfilter, filter_rows=newfilter_rows, filter_cols=newfilter_cols
        )

        # set to False so that totalsX has invalid where the badcols are
        # set to False so that totalsY has invalid where the badrows are
        # badcalc =True

        invalid = INVALID_DICT[imatrix.dtype.num]

        # if nothing is filtered, the calculation is simple
        if newfilter_rows is None and newfilter_cols is None:
            # calc total of totals - cell on far right and bottom
            totalOfTotals = func(totalsY)

        else:
            im = imatrix
            if not showfilter:
                # remove invalid row (columns in imatrix already removed)
                im = im[1:, :]

            # create row and col mask, init to all True
            boolmaskY = ones(len(totalsY), dtype=bool)
            boolmaskX = ones(len(totalsX), dtype=bool)

            # do both horizontal and vertical calculations which are dirty
            newTotalsX, newTotalsY = cls._apply_2d_operation(func, im, True)

            if newfilter_rows is not None:
                if showfilter:
                    # all the rows are shifted one over because Filtered row comes first
                    newfilter_rows += 1
                boolmaskY[newfilter_rows] = False

                # now set the invalids (hide the good value)
                # if not badcalc:
                #    totalsY[newfilter_rows] = invalid

                # shrink rows by removing bad values
                im = im[boolmaskY, :]

            if newfilter_cols is not None:
                if showfilter:
                    # all the cols are shifted one over because Filtered row comes first
                    newfilter_cols += 1
                boolmaskX[newfilter_cols] = False

                # now set the invalids (hide the good value)
                # if not badcalc:
                #    totalsX[newfilter_cols] = invalid

                # shrink cols
                im = im[:, boolmaskX]

            # do both horizontal and vertical calculations which are dirty
            newTotalsX, newTotalsY = cls._apply_2d_operation(func, im, True)

            if Accum2.DebugMode:
                numrows, numcols = imatrix.shape
                print("orig imatrix rows:", numrows, " cols:", numcols)
                numrows, numcols = im.shape
                print("new imatrix rows:", numrows, " cols:", numcols)

                print("oldtotalsX:", totalsX, " oldtotalsY:", totalsY)
                print("newtotalsX:", newTotalsX, " newtotalsY:", newTotalsY)

            # now repopulate with new calculation (with bad rows or cols removed)
            totalsY[boolmaskY] = newTotalsY
            totalsX[boolmaskX] = newTotalsX

            # if badcalc:
            #    if newfilter_rows is not None:
            #        if showfilter:
            #            # all the rows are shifted one over because Filtered row comes first
            #            newfilter_rows += 1
            #        boolmaskY[newfilter_rows] = False

            #        # shrink rows by removing bad values
            #        im = im[boolmaskY, :]

            #    if newfilter_cols is not None:
            #        if showfilter:
            #            # all the cols are shifted one over because Filtered row comes first
            #            newfilter_cols += 1
            #        boolmaskX[newfilter_cols] = False

            #        # shrink cols
            #        im = im[:,boolmaskX]

            # calc totals with both rows and cols removed
            totalOfTotals = func(im)

        # push calculations to dataset (newds)
        cls._add_totals(cat_rows, newds, name, totalsX, totalsY, totalOfTotals)

    # ---------------------------------------------------------------
    def _stack_dataset(self, arr, origarr, funcNum, showfilter: bool = False, tups=0, **kwargs):
        """
        Accum2 uses a single array but returns a dataset that is stacked.
        The long column is unrolled into columns.

        Parameters
        ----------
        arr :
        origarr :
        funcNum
        showfilter : bool
        kwargs : dict-like
            Keyword args to pass to the function specified by `funcNum`.

        """
        result = self.make_dataset(arr, showfilter=showfilter)

        newds = result["ds"]
        col_keys = result["col_keys"]
        row_keys = result["row_keys"]
        gbkeys = result["gbkeys"]

        # when user types in something like
        # ac.apply_reduce(lambda x,y:np.sum(np.maximum(x,y)), (newds.data, newds.data2))
        # we do not have the second parameter to pass in when we get to totals
        if self._totals and tups == 0:
            imatrix = self._make_imatrix(arr, col_keys, row_keys, showfilter)
            if not callable(funcNum):
                # get the name, func, routine to call
                func = apply_dict_total.get(funcNum)
                if func and callable(func[0]):
                    func_name = func[0](kwargs["func_param"])
                    func = (func_name, func[1], func[2])
            else:
                # assume _reduce
                # funcNum is really a callable function, we can get the name
                name = funcNum.__name__.capitalize()
                if name.startswith("<"):
                    name = "Lambda"

                func = (name, funcNum, Accum2._calc_multipass)

            if func is not None:
                # Calling single or multipass
                # may come from apply_reduce
                func[2](
                    self._cat_cols,
                    self._cat_rows,
                    newds,
                    origarr,
                    funcNum,
                    func[1],
                    imatrix,
                    name=func[0],
                    showfilter=showfilter,
                    **kwargs,
                )

        # tell display which columns are grouped by
        newds.label_set_names([k for k in gbkeys])

        # set any badrows/cols if we have them
        badrows = kwargs.get("badrows", None)
        if badrows is not None:
            # badrows wants fancy index
            badrows = Accum2._calc_badslots(self._cat_rows, badrows, None, True)
            if showfilter:
                badrows += 1
            newds._badrows = badrows

        badcols = kwargs.get("badcols", None)
        if badcols is not None:
            badcols = Accum2._calc_badslots(self._cat_cols, badcols, None, True)
            # badcols wants strings
            newds._badcols = self._cat_cols.category_array[badcols].astype("U")
            # print("set badcols to", newds._badcols, badcols)

        return newds

    # ---------------------------------------------------------------
    # OVERRIDEN from groupbyops
    def _calculate_all(self, funcNum, *args, func_param=0, **kwargs):
        """
        Can be called from apply_reduce

        """

        keychain = self.gb_keychain
        origdict, user_args, tups = self._prepare_gb_data("Accum2", funcNum, *args, **kwargs)

        # insert showfilter from Accum2 init if not overridden
        kwargs["showfilter"] = kwargs.get("showfilter", self._showfilter)

        if self._myfilter is not None:
            if kwargs.get("filter", None) is not None:
                kwargs["filter"] = self._myfilter & kwargs["filter"]
            else:
                kwargs["filter"] = self._myfilter

        if len(origdict) > 0:
            test_col = list(origdict.values())[0]
            if not isinstance(test_col, np.ndarray):
                # to get here pass in something like a lambda function as input
                raise ValueError(f"Data passed in to Accum2 must be numpy arrays and not type {type(test_col)!r}")

            if len(test_col.view(FastArray)) != len(self._cat_rows._fa):
                raise ValueError(
                    f"Data did not have the same length has categoricals in Accum2 object, {len(test_col.view(FastArray))} vs. {len(self._cat_rows._fa)}"
                )
        else:
            warnings.warn(f"Accum2: No data was calculated")
            return

        accum_dict = self.grouping._calculate_all(
            origdict,
            funcNum,
            func_param=func_param,
            keychain=keychain,
            user_args=user_args,
            tups=tups,
            return_all=self._return_all,
            accum2=True,
            **kwargs,
        )

        return self._finish_calculate_all(origdict, accum_dict, funcNum, func_param=func_param, tups=tups, **kwargs)

    # ---------------------------------------------------------------
    def _finish_calculate_all(self, origdict, accum_dict, funcNum, func_param=0, tups=0, transform=False, **kwargs):
        """

        Parameters
        ----------
        origdict: original dataset input
        accum_list: input data we can calculate on
        funcNum: internal riptable groupby function number  OR
                 a callable reduce function
        func_param: optional, parameters for the function
        """
        # check if transform was called earlier
        if transform:
            ikey = self.ikey
            # use fancy indexing to pull the values from the cells, back to original array
            newds = {colname: arr[ikey] for colname, arr in accum_dict.items()}

            return TypeRegister.Dataset(newds)

        accum_list = []
        for k, v in accum_dict.items():
            if isinstance(origdict[k], TypeRegister.Categorical):
                v = TypeRegister.Categorical(v, _from_categorical=origdict[k]._categories_wrap)
            accum_list.append(v)

        if Accum2.DebugMode:
            print("**accumlist was", accum_list)
            print("**origdict was", origdict)

        # if long mode
        if not callable(funcNum) and funcNum >= GB_FUNCTIONS.GB_CUMSUM:
            return accum_list

        # todo -- make ds columns?
        if len(accum_list) == 1:
            # just one input so normal dataset returned
            return self._stack_dataset(
                accum_list[0], origdict.popitem()[1], funcNum, func_param=func_param, tups=tups, **kwargs
            )

        elif len(accum_list) > 1:
            # multiple inputs, so use multiset
            # get the row and colkeys
            gbkeys = self._get_gbkeys()
            colkeys = self._cat_cols.grouping_dict

            ms = TypeRegister.Multiset({})
            for i, v in enumerate(accum_dict.keys()):

                # Accum2 uses two categoricals for rows and cols
                # dont bother to calculate the row or col keys
                # the row keys are the gbkeys
                if v not in gbkeys and v not in colkeys:
                    result = self._stack_dataset(
                        accum_list[i], origdict[v], funcNum, func_param=func_param, tups=tups, **kwargs
                    )
                    ms[v] = result

            gbkeyname = self._get_gbkeyname()
            ms._gbkeys = gbkeys
            # ms.label_set_names([gbkeyname])
            return ms
        else:
            print("No data was calculated")
            return None

    # ---------------------------------------------------------------
    def apply_reduce(
        self, userfunc, *args, dataset=None, label_keys=None, func_param=None, dtype=None, transform=False, **kwargs
    ):

        """
        Accum2:apply_reduce calls Grouping:apply_helper

        Parameters
        ----------
        userfunc : callable
            A callable that takes a contiguous array as its first argument, and returns a scalar
            In addition the callable may take positional and keyword arguments.
        args
            Used to pass in columnar data from other datasets

        Other Parameters
        ----------------
        dataset: None
            User may pass in an entire dataset to compute.
        label_keys: None
            Not supported, will use the existing groupby keys as labels.
        func_param : tuple, optional
            Set to a tuple to pass as arguments to the routine.
        dtype : str or np.dtype, optional
            Change to a numpy dtype to return an array with that dtype. Defaults to None.
        transform : bool
            Set to True to re-expand the results of the calculation. Defaults to False.
        filter:
        kwargs
            Optional positional and keyword arguments to pass to ``userfunc``

        Notes
        -----
        See Grouping.apply_reduce
        """

        args, kwargs, origdict, tups = self._pop_gb_data("apply_reduce", userfunc, *args, **kwargs, dataset=dataset)
        # temp pop showfilter
        realshowfilter = kwargs.get("showfilter", False)
        kwargs["showfilter"] = True

        accum_dict = self.grouping.apply_helper(
            True, origdict, userfunc, *args, tups=tups, label_keys=None, func_param=func_param, dtype=dtype, **kwargs
        )
        kwargs["showfilter"] = realshowfilter

        return self._finish_calculate_all(
            origdict, accum_dict, userfunc, func_param=func_param, tups=tups, transform=transform, **kwargs
        )

    # ---------------------------------------------------------------
    @property
    def ncountkey(self):
        """See: Grouping.ncountkey"""
        return self.grouping.ncountkey

    # ---------------------------------------------------------------
    @property
    def ncountgroup(self):
        """See: Grouping.ncountgroup"""
        return self.grouping.ncountgroup

    # ---------------------------------------------------------------
    @property
    def ikey(self):
        return self.grouping.ikey

    # ---------------------------------------------------------------
    def count(self, **kwargs):
        """Compute count of group"""
        kwargs["showfilter"] = kwargs.get("showfilter", self._showfilter)
        # print("count array", self.ncountgroup)
        result = self._stack_dataset(self.ncountgroup, self.ncountkey, GB_FUNC_COUNT, **kwargs)
        if self._totals:
            result = result.imatrix_totals()
        return result

    # -------------------------------------------------------
    def _build_string(self):
        # build a count dataset by default
        dset = self.count()
        resultString = dset.__str__()
        return (
            "Accum2 Keys\n X:"
            + self._cat_cols.unique_repr
            + "\n Y:"
            + self._cat_rows.unique_repr
            + "\n Bins:"
            + str(self.grouping.unique_count)
            + "   Rows:"
            + str(len(self._cat_rows._np))
            + "\n\n"
            + resultString
        )

    # -------------------------------------------------------
    def _build_sds_meta_data(self, name, **kwargs):
        meta = MetaData(
            {
                # vars for container loader
                "name": name,
                "typeid": TypeId.Accum2,
                "classname": self.__class__.__name__,
                "_base_is_stackable": 0,
                "author": "python",
                # accum2 will always have 2 categoricals for x and y axis
                "cat_meta": [],
                "ncols": 0,  # one for each cat underlying array
                # what are these
                "instance_vars": {
                    "_showfilter": self._showfilter,
                },
            }
        )

        cols = []

        for cat in [self._cat_cols, self._cat_rows]:
            cols.append(cat._fa)
            cat_meta, cat_cols = cat._build_sds_meta_data(TypeId.Categorical.name)
            for c in cat_cols:
                cols.append(c)
            meta["cat_meta"].append(cat_meta.string)

        meta["ncols"] = len(cols)

        return meta, cols

    # -------------------------------------------------------
    @classmethod
    def _load_from_sds_meta_data(self, name, arr, cols, meta):
        if not isinstance(meta, MetaData):
            meta = MetaData(meta)

        # what instance vars need to be stored
        vars = meta["instance_vars"]

        col_idx = 0
        cats = []
        # build categoricals for x and y
        for cat_meta in meta["cat_meta"]:
            cat_meta = MetaData(cat_meta)
            # first array in block is categoricals underlying
            cat_arr = cols[col_idx]
            col_idx += 1

            # next columns are for categories
            cat_cols = cols[col_idx : col_idx + cat_meta["ncols"]]
            col_idx += cat_meta["ncols"]

            # ask categorical to reconstruct
            cat = TypeRegister.Categorical._load_from_sds_meta_data(cat_meta["name"], cat_arr, cat_cols, cat_meta)
            cats.append(cat)

        # TODO: add a fast-track routine to reconstruct accum2 object with cats, underlying array
        return Accum2(cats[0], cats[1], showfilter=vars["_showfilter"])

    # -------------------------------------------------------
    def __repr__(self):
        return self._build_string()

    # -------------------------------------------------------
    def __str__(self):
        return self._build_string()


# mapping of internal functions to the proper np routine
# for sum we only need to make one pass (calc_onepass)
# for mean we need to make multiple passes
apply_dict_total = {
    GB_FUNCTIONS.GB_SUM: ("Total", np.sum, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_NANSUM: ("Nansum", np.nansum, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_MIN: ("Min", np.min, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_NANMIN: ("Nanmin", np.nanmin, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_MAX: ("Max", np.max, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_NANMAX: ("Nanmax", np.nanmax, Accum2._calc_onepass),
    GB_FUNCTIONS.GB_MEAN: ("Mean", np.mean, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_NANMEAN: ("Nanmean", np.nanmean, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_VAR: ("Var", np.var, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_NANVAR: ("Nanvar", np.nanvar, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_STD: ("Std", np.std, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_NANSTD: ("Nanstd", np.nanstd, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_MEDIAN: ("Median", np.nanmedian, Accum2._calc_multipass),
    # Didn't change name of Median to Nanmedian in GroupByOps.quantile_name_from_param to avoid breaking existing things
    # quantile functions take an argument with 1e9 multiplier, so use simple wrappers in GroupByOps
    GB_FUNCTIONS.GB_QUANTILE_MULT: (
        GroupByOps.quantile_name_from_param,
        GroupByOps.np_quantile_mult,
        Accum2._calc_multipass,
    ),
    GB_FUNCTIONS.GB_MODE: ("Mode", None, Accum2._calc_multipass),
    GB_FUNCTIONS.GB_TRIMBR: ("Trimbr", None, Accum2._calc_multipass),
    # -- transform (or same size) functions
    # GB_FUNCTIONS.GB_CUMSUM : ('Cumsum', np.cumsum, Accum2._calc_multipass),
}


# keep this as the last line
TypeRegister.Accum2 = Accum2



__all__ = ["TestData", "load_test_data"]

import numpy as np

from ..rt_enum import NumpyCharTypes
from ..rt_enum import TypeRegister as treg
from ..rt_numpy import *
from ..rt_timers import tt


class TestDataMeta(type):
    length = 20

    _dataset_numeric = None
    _dataset_strings = None
    _dataset_rt_types = None
    _dataset_for_gb = None
    _categorical_all = None
    _struct_nested = None

    _categorical_bytes = None
    _categorical_numeric = None
    _categorical_codes = None
    _categorical_mapping = None
    _categorical_multikey = None
    _categorical_matlab = None
    _categorical_pandas = None

    _categorical_firstorder = None
    _categorical_unicode = None
    _categorical_zero = None

    _int_fa = None
    _bytes_fa = None
    _unicode_fa = None
    _flt_fa = None
    _flt_fa2 = None
    _str_map = None

    _rand_bool = None
    # ------------------DATASETS-------------------------------------------
    # ---------------------------------------------------------------------
    @property
    def dataset_numeric(cls):
        # possibly reset dataset with new length
        if cls._dataset_numeric is not None and cls._dataset_numeric._nrows != cls.length:
            cls._dataset_numeric = None

        if cls._dataset_numeric is None:
            cls._dataset_numeric = treg.Dataset(
                {
                    "int8": np.random.randint(low=-128, high=127, size=(cls.length), dtype=np.int8),
                    "uint8": np.random.randint(low=0, high=255, size=(cls.length), dtype=np.uint8),
                    "int16": np.random.randint(low=-32768, high=32767, size=(cls.length), dtype=np.int16),
                    "uint16": np.random.randint(low=0, high=65535, size=(cls.length), dtype=np.uint16),
                    "int32": np.random.randint(
                        low=-2147483648,
                        high=2147483647,
                        size=(cls.length),
                        dtype=np.int32,
                    ),
                    "uint32": np.random.randint(low=0, high=4294967295, size=(cls.length), dtype=np.uint32),
                    "int64": np.random.randint(
                        low=-9223372036854775808,
                        high=9223372036854775807,
                        size=(cls.length),
                        dtype=np.int64,
                    ),
                    "uint64": np.random.randint(
                        low=0,
                        high=18446744073709551615,
                        size=(cls.length),
                        dtype=np.uint64,
                    ),
                    "float32": np.random.rand(cls.length).astype(np.float32) * 10e9,
                    "float64": np.random.rand(cls.length).astype(np.float64) * 10e-9,
                }
            )
        return cls._dataset_numeric

    # ---------------------------------------------------------------------
    @property
    def dataset_strings(cls):
        # possibly reset dataset with new length
        if cls._dataset_strings is not None and cls._dataset_strings._nrows != cls.length:
            cls._dataset_strings = None

        if cls._dataset_strings is None:
            cls._dataset_strings = treg.Dataset(
                {
                    "unicode": treg.FastArray(
                        np.random.choice(
                            [
                                "peter",
                                "paul",
                                "mary ann",
                                "joe",
                                "john",
                                "ray",
                                "mary katherine",
                            ],
                            cls.length,
                        ),
                        unicode=True,
                    ),
                    "bytes": treg.FastArray(
                        np.random.choice(
                            ["greg", "julia", "paul", "marc", "chuck", "mary"],
                            cls.length,
                        )
                    ),
                    "categorical": treg.Categorical(
                        np.random.randint(1, 8, cls.length),
                        ["mary ann", "paul", "sam", "charlie", "norm", "lenny", "tula"],
                    ),
                }
            )
        return cls._dataset_strings

    # ---------------------------------------------------------------------
    @property
    def dataset_for_gb(cls):
        """
        This is a good one for testing groupby results/sorting/apply/etc.
        """
        if cls._dataset_for_gb is not None and cls._dataset_for_gb._nrows != cls.length:
            cls._dataset_for_gb = None

        if cls._dataset_for_gb is None:
            cls._dataset_for_gb = treg.Dataset({"col_" + str(i): np.random.rand(cls.length) for i in range(10)})
            cls._dataset_for_gb.keycol = treg.FastArray(
                np.random.choice(
                    [
                        "peter",
                        "paul",
                        "mary ann",
                        "joe",
                        "john",
                        "ray",
                        "mary katherine",
                    ],
                    cls.length,
                ),
                unicode=True,
            )

        return cls._dataset_for_gb

    # ---------------------------------------------------------------------
    @property
    def dataset_rt_types(cls):
        if cls._dataset_rt_types is None:
            cls._dataset_rt_types = treg.Dataset(
                {
                    "CAT": cls.categorical_bytes,
                    "DTN": treg.DateTimeNano.random(cls.length),
                    "DATE": treg.Date(np.random.randint(15000, 20000, cls.length)),
                    "TSPAN": treg.TimeSpan(
                        np.random.randint(0, 1_000_000_000 * 60 * 60 * 24, cls.length, dtype=np.int64)
                    ),
                    "DSPAN": treg.DateSpan(np.random.randint(0, 365, cls.length)),
                }
            )
        return cls._dataset_rt_types

    # -----STRUCTS---------------------------------------------------------
    # ---------------------------------------------------------------------
    @property
    def struct_nested(cls):
        if cls._struct_nested is None:
            cls._struct_nested = treg.Struct(
                {
                    "cat1": treg.Categorical(["a", "b", "c", "d"]),
                    "arr1": arange(5),
                    "ds1": treg.Dataset({"col_" + str(i): np.random.rand(5) for i in range(5)}),
                    "arr2": arange(10),
                    "test_string": "this is my string",
                    "array3": arange(30).astype(np.int8),
                    "nested_struct1": treg.Struct(
                        {
                            "set1": treg.Dataset({"col_" + str(i): np.random.randint(5) for i in range(3)}),
                            "set2": treg.Dataset(
                                {"col_" + str(i): np.random.choice(["a", "b", "c"], 10) for i in range(2)}
                            ),
                            "nested_2": treg.Struct(
                                {
                                    "c1": treg.Categorical(arange(5)),
                                    "c2": treg.Categorical(["aaa", "bbbb", "cccc"]),
                                    "leaf_dataset": treg.Dataset({"col_" + str(i): arange(5) for i in range(3)}),
                                    "array4": arange(20).astype(np.uint64),
                                }
                            ),
                            "set3": treg.Dataset(
                                {"col_" + str(i): np.random.choice([True, False], 10) for i in range(4)}
                            ),
                        }
                    ),
                    "ds2": treg.Dataset({"heading_" + str(i): np.random.rand(5) for i in range(3)}),
                    "int1": 5,
                    "float1": 7.0,
                    "cat2": treg.Categorical(["a", "b", "c", "d"]),
                }
            )
        return cls._struct_nested

    # -----CATEGORICALS----------------------------------------------------
    # ---------------------------------------------------------------------
    def _gen_categorical_data(cls):
        if cls._int_fa is not None and len(cls._int_fa) != cls.length:
            cls._int_fa = None

        if cls._int_fa is None:
            cls._int_fa = treg.FastArray(np.random.choice([10, 20, 30], cls.length))
            cls._bytes_fa = treg.FastArray(
                np.random.choice(["adam", "brian", "charlie", "david", "edward"], cls.length)
            )
            cls._unicode_fa = treg.FastArray(
                np.random.choice(["amanda", "beth", "catherine", "danielle", "emma"], cls.length),
                unicode=True,
            )
            cls._flt_fa = treg.FastArray(np.random.rand(cls.length)) * 100
            cls._flt_fa2 = treg.FastArray(np.random.rand(cls.length)) * 100
            cls._str_map = {10: "x", 20: "y", 30: "z", 40: "zz"}

    def _gen_categorical(cls, name):
        cat = getattr(cls, name)
        if cat is not None and len(cat) != cls.length:
            cat = None

        if cat is None:
            cls._gen_categorical_data()

        return cat

    @property
    def categorical_all(cls):
        data = treg.Dataset(
            {
                "categorical_bytes": cls.categorical_bytes,
                "categorical_unicode": cls.categorical_unicode,
                "categorical_multikey": cls.categorical_multikey,
                "categorical_firstorder": cls.categorical_firstorder,
                "categorical_zero": cls.categorical_zero,
                "categorical_numeric": cls.categorical_numeric,
            }
        )
        return data

    @property
    def categorical_bytes(cls):
        cat = cls._gen_categorical("_categorical_bytes")
        if cat is None:
            cls._categorical_bytes = treg.Categorical(cls._bytes_fa)
        return cls._categorical_bytes

    @property
    def categorical_unicode(cls):
        cat = cls._gen_categorical("_categorical_unicode")
        if cat is None:
            cls._categorical_unicode = treg.Categorical(cls._unicode_fa, unicode=True)
        return cls._categorical_unicode

    @property
    def categorical_numeric(cls):
        cat = cls._gen_categorical("_categorical_numeric")
        if cat is None:
            cls._categorical_numeric = treg.Categorical(cls._int_fa)
        return cls._categorical_numeric

    @property
    def categorical_multikey(cls):
        cat = cls._gen_categorical("_categorical_multikey")
        if cat is None:
            cls._categorical_multikey = treg.Categorical([cls._bytes_fa, cls._int_fa])
        return cls._categorical_multikey

    @property
    def categorical_firstorder(cls):
        cat = cls._gen_categorical("_categorical_firstorder")
        if cat is None:
            cls._categorical_firstorder = treg.Categorical(cls._bytes_fa, ordered=False)
        return cls._categorical_firstorder

    @property
    def categorical_zero(cls):
        cat = cls._gen_categorical("_categorical_zero")
        if cat is None:
            cls._categorical_zero = treg.Categorical(cls._bytes_fa, base_index=0)
        return cls._categorical_zero

    # ----ARRAYS-----------------------------------------------------------
    # ---------------------------------------------------------------------
    @property
    def rand_bool(cls):
        if cls._rand_bool is not None and len(cls._rand_bool) != cls.length:
            cls._rand_bool = None
        if cls._rand_bool is None:
            cls._rand_bool = np.random.randint(0, 2, cls.length, dtype=bool).view(treg.FastArray)
        return cls._rand_bool

    # ---SETTINGS----------------------------------------------------------
    # ---------------------------------------------------------------------
    def reset(cls):
        cls._dataset_numeric = None
        cls._dataset_strings = None
        cls._dataset_rt_types = None
        cls._dataset_for_gb = None
        cls._categorical_all = None
        cls._struct_nested = None

        cls._categorical_bytes = None
        cls._categorical_numeric = None
        cls._categorical_codes = None
        cls._categorical_mapping = None
        cls._categorical_multikey = None
        cls._categorical_matlab = None
        cls._categorical_pandas = None

        cls._categorical_firstorder = None
        cls._categorical_unicode = None
        cls._categorical_zero = None

        _rand_bool = None

        cls._int_fa = None
        cls._bytes_fa = None
        cls._unicode_fa = None
        cls._flt_fa = None
        cls._flt_fa2 = None
        cls._str_map = None

    def set_array_size(cls, sz):
        if not isinstance(sz, (int, np.integer)):
            raise TypeError(f"Array length must be an integer, not {type(size)}")
        cls.length = sz

    # ----PERFORMANCE TESTS------------------------------------------------
    # ---------------------------------------------------------------------
    def _test_arr_from_dt(cls, nrows=1_000_000, dtype=np.dtype(np.float64)):
        if dtype.char in NumpyCharTypes.AllFloat:
            arr = np.asarray(np.random.rand(nrows), dtype=dtype)
        elif dtype.char in NumpyCharTypes.AllInteger:
            maxint = np.iinfo(dtype).max
            maxint = min(nrows, maxint)
            arr = np.random.randint(0, maxint, size=nrows, dtype=dtype)
        else:
            raise ValueError(
                f"Cannot perform reduce test with {dtype} dtype. Use integer or floating point type instead."
            )

        return arr

    # ---------------------------------------------------------------------
    def test_dataset_reduce(
        cls,
        nrows=1_000_000,
        ncols=10,
        dtype=np.float64,
        loops=2,
        transpose=False,
        apply=False,
    ):
        reduce_funcs = [
            "sum",
            "nansum",
            "mean",
            "nanmean",
            "std",
            "nanstd",
            "var",
            "nanvar",
            "min",
            "nanmin",
            "max",
            "nanmax",
        ]

        import pandas as pd

        if dtype is None:
            dtype = list(NumpyCharTypes.Computable)

        if not isinstance(dtype, list):
            dtype = [dtype]

        arrs = []
        if isinstance(dtype, list):
            dts = []
            # validate all dtypes
            for item in dtype:
                try:
                    dts.append(np.dtype(item))
                except:
                    raise ValueError(f"{item} was not a valid numpy dtype initializer.")

            # generate arrays
            for i in range(ncols):
                dt = np.random.choice(dts)
                arrs.append(cls._test_arr_from_dt(nrows, dt))

        ds = treg.Dataset({"c" + str(i): arrs[i] for i in range(ncols)})
        df = pd.DataFrame(ds.asdict())

        rt_times = []
        pd_times = []

        if apply:
            for i, func_str in enumerate(reduce_funcs):
                print(f"testing apply {func_str}")
                rt_str = f"ds.apply({func_str})"
                pd_str = f"df.apply(np.{func_str})"

                rt_times.append(tt(rt_str, loops=loops, return_time=True))
                pd_times.append(tt(pd_str, loops=loops, return_time=True))
        else:
            for i, func_str in enumerate(reduce_funcs):
                print(f"testing {func_str}")
                rt_str = f"ds.{func_str}()"
                if i % 2 == 1:
                    pd_str = f"df.{reduce_funcs[i-1]}()"
                else:
                    pd_str = f"df.{func_str}(skipna=False)"

                rt_times.append(tt(rt_str, loops=loops, return_time=True))
                pd_times.append(tt(pd_str, loops=loops, return_time=True))

        rt_times = treg.FastArray(rt_times)
        pd_times = treg.FastArray(pd_times)
        speedup = pd_times / rt_times

        result = treg.Dataset(
            {
                "operation": treg.FastArray(reduce_funcs),
                "pandas time (s)": pd_times,
                "riptable time (s)": rt_times,
                "speedup": speedup,
            }
        )

        result.sort_inplace("speedup", ascending=False)
        result.speedup = treg.FastArray(["{0:.2f}".format(i) + "x" for i in result.speedup])

        return result._temp_display("PRECISION", 3)


class TestData(metaclass=TestDataMeta):
    pass


def load_test_data(generate=True):
    """
    Return a Struct containing all of the test data in the TestData class (will all be generated if necessary)
    """
    data = treg.Struct(
        {
            # FastArray test data
            "fastarray_bool": TestData.rand_bool,
            # Categorical test data
            # 'categorical_all': TestData.categorical_all,
            "categorical_bytes": TestData.categorical_bytes,
            "categorical_firstorder": TestData.categorical_firstorder,
            "categorical_multikey": TestData.categorical_multikey,
            "categorical_numeric": TestData.categorical_numeric,
            "categorical_unicode": TestData.categorical_unicode,
            "categorical_zero": TestData.categorical_zero,
            # Dataset test data
            "dataset_numeric": TestData.dataset_numeric,
            "dataset_strings": TestData.dataset_strings,
            "dataset_rt_types": TestData.dataset_rt_types,
            "dataset_for_gb": TestData.dataset_for_gb,
            # Struct test data
            # TODO enable struct_nested after fixing Struct.equality check
            # 'struct_nested': TestData.struct_nested,
        }
    )
    return data



__all__ = ["AccumTable", "accum_ratio", "accum_ratiop", "accum_cols"]


import warnings
from collections import OrderedDict

import numpy as np

from .rt_accum2 import Accum2
from .rt_categorical import Categorical
from .rt_enum import TypeRegister
from .rt_numpy import full


class AccumTable(Accum2):
    """
    AccumTable is a wrapper on Accum2 that enables the creation of tables that
    combine the results of multiple reductions generated from the Accum2 object.
    The three parts of a table generated by the AccumTable gen() method are these:

    * **Inner Table** - a table of values indexed by row labels and column key names. A
      generated table contains only one inner table, but any number of inner tables may be created
      and used to create margin columns and footer rows, or as a reference for display formatting
      (future functionality).

    * **Margin Columns** - columns of values on the right margin, associated with an inner table, and
      indexed by and representing a value associated with a given row label.  Call
      set_margin_columns() to adjust them.

    * **Footer Rows** - rows of values on the bottom margin, associated with an inner table and indexed by
      and representing a value associated with a given column key.  Call
      set_footer_rows() to adjust them.

    Parameters
    ----------
    cat_rows: Categorical or an array converted to same
        The array used to create the row labels in the AccumTable
    cat_cols: Categorical or an array converted to same
        The array used to create the column keys in the AccumTable
    filter: ndarray
        Boolean mask array applied as filter before constructing the groupings
    showfilter: bool
        Whether to include groupings whose values were all filtered out
    """

    # -------------------------------------------------------
    def __init__(cls, cat_rows, cat_cols, filter=None, showfilter=False):
        pass

    def __new__(cls, cat_rows, cat_cols, filter=None, showfilter=False):
        """

        Parameters
        ----------
        cat_rows: Categorical or an array converted to same
            The array used to create the row labels in the AccumTable
        cat_cols: Categorical or an array converted to same
            The array used to create the column keys in the AccumTable
        filter: ndarray
            Boolean mask array applied as filter before constructing the groupings
        showfilter: bool
            Whether to include groupings whose values were all filtered out

        Returns
        -------
        AccumTable
            The new instance
        """
        instance = super(AccumTable, cls).__new__(cls, cat_rows, cat_cols, filter, showfilter)
        instance._inner = OrderedDict()
        instance._rows = OrderedDict()
        instance._cols = OrderedDict()
        instance._default_inner_name = None
        return instance

    # -------------------------------------------------------
    def __repr__(self):
        """
        Return a string representation of the object

        Returns
        -------
        str
            The repr string
        """
        res = "Inner Tables: " + str(list(self._inner.keys())) + "\n"
        res += "Margin Columns: " + str(list(self._cols.keys())) + "\n"
        res += "Footer Rows: " + str(list(self._rows.keys()))
        return res

    # -------------------------------------------------------
    def __setitem__(self, name: str, ds):
        """
        Parameters
        ----------
        name : str
            Inner table name
        ds : Dataset
            The dataset

        Raises
        ------
        IndexError
            If `name` is not a string (table name).
        ValueError
            If `ds` is not a Dataset
        """
        if not type(name) is str:
            raise IndexError("name must be a string table name")
        if not isinstance(ds, TypeRegister.Dataset):
            raise ValueError("ds must be a Dataset")
        self._inner[name] = ds
        self._rows[name] = None
        self._cols[name] = None
        self._rename_summary_row_and_col(ds, name)
        self._default_inner_name = name

    # -------------------------------------------------------
    def __getitem__(self, index: str):
        """
        Parameters
        ----------
        index : str
            Inner table name

        Returns
        -------
        Dataset
            The specified inner table

        Raises
        ------
        IndexError
            If `index` is not a string (table name).
        """
        if not type(index) is str:
            raise IndexError("Index must be a string table name")
        self._default_inner_name = index
        return self._inner[index]

    # -------------------------------------------------------
    def _rename_summary_row_and_col(self, ds, new_name: str):
        """
        Parameters
        ----------
        ds : Dataset
            The dataset
        new_name : str
            the new name for the summary column and footer row

        Returns
        -------
        Dataset
        """
        col_names = ds.summary_get_names()
        if len(col_names) == 1:
            ds.col_rename(col_names[0], new_name)
        footers = ds.footer_get_dict()
        if len(footers) == 1:
            old_name = list(footers.keys())[0]
            nd = list(footers.values())[0]
            ds.footer_remove(old_name)
            ds.footer_set_values(new_name, nd)
        return ds

    # -------------------------------------------------------
    def gen(self, table_name=None, format=None, ref_table=None, remove_blanks=True):
        """
        Generate an AccumTable view.

        Parameters
        ----------
        table_name: string or tuple (not implemented yet)
            The name of the AccumTable table to display, or a tuple of table names if
            more than one value is to be displayed in each cell (not implemented yet).
        format (not implemented yet): dict
            A dictionary used to specify the formatting of each cell in the table. The
            keys are formatting types, such as 'bold', 'color', and 'background',
            and the values are functions that are applied to the value (or tuple)
            in each table cell to determine the applicability of a formatting type.
            For example, once could set format={'bold': lambda v: v > 0} to make all
            positive values in the table bold.
        ref_table (not implemented yet): string or Dataset
            The name of the AccumTable table, or a Dataset of the same shape, to be
            referenced for formatting the displayed table (not implemented yet).
        remove_blanks: bool
            Do not display rows or columns containing all zeros or NaNs

        Returns
        -------
        Dataset
            The generated table

        Examples
        --------
        View the pnl values in the table, coloring negative values red (not implemented yet):

        >>> at.gen('pnl', format={'color': lambda v: return 'red' if v < 0 else 'black'})
        """
        # Get the displayed, inner table
        table_name = self._default_inner_name if table_name is None else table_name
        self._default_inner_name = table_name
        if table_name is None:
            raise ValueError("Must specify a table name")
        orig = self._inner[table_name]

        # Remove blanks, as required, and set the row filter
        if remove_blanks:
            (clean, row_filter, _) = orig.copy().trim(ret_filters=True)
            row_filter = row_filter if row_filter is not None else slice(None, None, None)
        else:
            clean = orig.copy()
            row_filter = slice(None, None, None)

        # Add the margin columns to the right
        summary_names = clean.summary_get_names()
        for mar_col in [col for col in list(self._cols.keys()) if col != table_name]:
            clean[mar_col] = self._inner[mar_col][row_filter, mar_col]
            summary_names += [mar_col]
        clean.summary_set_names(summary_names)

        # Add the footer rows at the bottom
        for footer_row in [row for row in list(self._rows.keys()) if row != table_name]:
            fd = list(self._inner[footer_row].footer_get_dict(footer_row).values())[0]
            delete = [k for k in fd.keys() if not k in clean.keys()]
            for key in delete:
                del fd[key]
            clean.footer_set_values(footer_row, fd)

        return clean

    # -------------------------------------------------------
    def set_margin_columns(self, cols):
        """
        Specify the names of the inner tables whose margin columns should appear in the generated AccumTable view.

        Parameters
        ----------
        cols: list of str
            The list of inner table names, in order.
        """
        self._cols = OrderedDict()
        for k in cols:
            self._cols[k] = None

    # -------------------------------------------------------
    def set_footer_rows(self, rows):
        """
        Specify the names of the inner tables whose footer rows should appear in the generated AccumTable view.

        Parameters
        ----------
        rows: list
            The list of inner table names, in order.
        """
        self._rows = OrderedDict()
        for k in rows:
            self._rows[k] = None


def accum_ratio(
    cat1,
    cat2=None,
    val1=None,
    val2=None,
    filt1=None,
    filt2=None,
    func1="nansum",
    func2=None,
    return_table=False,
    include_numer=False,
    include_denom=True,
    remove_blanks=True,
):
    """
    Compute a bucketed ratio of two accums, using AccumTable.

    Parameters
    ----------
    cat1 : Categorical
        First categorical label to group by.
    cat2 : Categorical, optional
        Second categorical label to group by.
    val1
        Numerator data
    val2
        Denominator data
    filt1
        Filter for `val1` data
    filt2
        Filter for `val2` data. Optional, defaults to `filter1`
    func1
        String of function name to pass into numerator AccumTable call
    func2
        String of function name to pass into denominator AccumTable call. Defaults to `func1`.
    include_numer : bool
        If set to True, include the totals from the numerator data in the output. Ignored if `return_table` is True.
    include_denom : bool
        If set to True, include the totals from the denominator data in the output. Ignored if `return_table` is True.
    return_table : bool
        If set to True, returns the whole AccumTable instead of just the gen'd ratio data.
    remove_blanks : bool
        If set to True, blanks will be removed from the output.

    Returns
    -------
    Dataset or AccumTable
        Either a view of the ratio data, or the entire AccumTable, depending on `return_table` flag.
    """
    # Handle missing inputs
    if val1 is None:
        raise ValueError("Missing argument val1")
    if (
        (val2 is None) & (cat2 is not None) & (val1 is not None)
    ):  # Passing as accum_ratio(cat1, val1, val2), omitting cat2 argument
        val2 = val1
        val1 = cat2
        cat2 = None
    if filt1 is None:
        filt1 = full(val1.shape[0], True, dtype=bool)  # This was playa.utils.truecol
    if filt2 is None:
        filt2 = filt1
    if func2 is None:
        func2 = func1
    if cat2 is None:
        cat2 = Categorical(full(val1.shape[0], 1, dtype=np.int8), ["NotGrouped"])  # This was playa.utils.onescol

    # Handle name collisions
    for key in ["Numer", "Denom", "Ratio"]:
        if key in cat2.categories():
            cat2.category_replace(key, key + "_")

    # Compute accum
    accum = AccumTable(cat1, cat2)

    func1 = getattr(accum, func1)
    func2 = getattr(accum, func2)
    # TODO: In the future, when arbitrary functions are allowed in Accum2 calls, handle a missing attr here by passing it in by name
    accum["Numer"] = func1(val1, filter=filt1)
    accum["Denom"] = func2(val2, filter=filt2)

    accum["Ratio"] = accum["Numer"] / accum["Denom"]

    if return_table:
        return accum
    else:
        footers = [label for (label, boolean) in zip(["Numer", "Denom"], [include_numer, include_denom]) if boolean]
        accum.set_margin_columns(footers)
        accum.set_footer_rows(footers)
        return accum.gen("Ratio", remove_blanks=remove_blanks)


def accum_ratiop(
    cat1,
    cat2=None,
    val=None,
    filter=None,
    func="nansum",
    norm_by="T",
    include_total=True,
    remove_blanks=True,
    filt=None,
):
    """
    Compute an internal ratio, either by Total (T), Row (R), or Column (C).

    Parameters
    ----------
    cat1 : Categorical
        First categorical label to group by
    cat2 : Categorical, optional
        Second categorical label to group by.
    val
        Data column
    filter: boolean column
        Filter for var data. Replacing filt.
    func : str
        String of function name to pass into AccumTable call
    norm_by : {'T', 'C', 'R'}
        what to use as the denominator
    include_total : bool
        Include the total amounts in addition to the ratios, defaults to True.
    remove_blanks : bool
        If set to True, blanks will be removed from the output; defaults to True.
    filt
        DEPRECATED FOR "filter".


    Returns
    -------
    Dataset
        AccumTable view of ratios
    """
    # Handle missing inputs
    if val is None:
        val = full(cat1.shape[0], 1, dtype=np.float64)  # This was playa.utils.onescol
    if filter is None:
        if filt is not None:  # Temporary until deprecated
            warnings.warn(
                'Kwarg "filt" is being deprecated for "filter" to align with common syntax. "filt" will be removed in a future version',
                FutureWarning,
            )
            filter = filt
        else:
            filter = full(val.shape[0], True, dtype=bool)  # This was playa.utils.truecol
    if cat2 is None:
        cat2 = Categorical(full(val.shape[0], 1, dtype=np.int8), ["NotGrouped"])  # This was playa.utils.onescol

    # Compute accum
    accum = AccumTable(cat1, cat2)

    func_name = func
    func = getattr(accum, func_name)
    # TODO: In the future, when arbitrary functions are allowed in Accum2 calls, handle a missing attr here by passing it in by name
    accum["TotalRatio"] = func(val, filter=filter)
    if include_total:
        accum["Total"] = func(val, filter=filter)

    accumr = accum.gen("TotalRatio", remove_blanks=remove_blanks)
    if include_total:
        keys = accumr.keys()[1:-1]
    else:
        keys = accumr.keys()[1:]

    if norm_by.upper() == "T":
        total = accumr.footer_get_dict()["TotalRatio"]["TotalRatio"]
        accumr.footer_set_values(
            "TotalRatio", {key: 100 * item / total for (key, item) in accumr.footer_get_dict()["TotalRatio"].items()}
        )
        for col in keys:
            accumr[col] = 100 * accumr[col] / total
    elif norm_by.upper() == "R":
        total = accumr.footer_get_dict()["TotalRatio"]["TotalRatio"]
        accumr.footer_set_values(
            "TotalRatio", {key: 100 * item / total for (key, item) in accumr.footer_get_dict()["TotalRatio"].items()}
        )
        for col in keys:
            accumr[col] = 100 * accumr[col] / accumr.TotalRatio
    elif norm_by.upper() == "C":
        for col in keys:
            total = accumr.footer_get_dict()["TotalRatio"][col]
            accumr[col] = 100 * accumr[col] / total
        accumr.footer_set_values(
            "TotalRatio", {key: 100.0 for (key, item) in accumr.footer_get_dict()["TotalRatio"].items()}
        )
    else:
        raise ValueError(f"Invalid norm_by selection: {norm_by}. Valid choices are T, R, C.")

    return accumr


def accum_cols(cat, val_list, name_list=None, filt_list=None, func_list="nansum", remove_blanks=True):
    """
    Compute multiple accum calculations on the same categorical label, output as a single dataset.

    Parameters
    ----------
    cat : Categorical
        Categorical label to group by
    val_list : list
        List of data columns. If an element is a two-element list itself, a ratio will be calculated.
        If an element is a two-element list of type [val, 'p'], an accum_ratiop-style percentile will be calculated
    name_list : list
        List of column names in the eventual dataset. Defaults to colN.
    filt_list : list
        List of filters, either one for all or one for each. Defaults to truecol.
    func_list : str or list of str
        String of function name (or list of strings of function names) to pass into AccumTable call,
        either one for all or one for each. Defaults to 'nansum'.
    remove_blanks : bool
        If set to true, blanks will be removed from the output. Defaults to True.

    Returns
    -------
    Dataset
        Accum2 view of calculated data.
    """

    # Handle mistyped inputs
    if not isinstance(cat, Categorical):
        cat = Categorical(cat)
    if not isinstance(val_list, list):
        val_list = [val_list]

    # Handle missing inputs
    if name_list is None:
        name_list = [f"col{n}" for n in range(len(val_list))]
    if filt_list is None:
        filt_list = full(val_list[0].shape[0], True, dtype=bool)  # This was playa.utils.truecol
    if not isinstance(func_list, list):
        func_list = [func_list for _ in val_list]
    if not isinstance(filt_list, list):
        filt_list = [filt_list for _ in val_list]

    # Compute accum
    temp_cat = Categorical(full(cat.shape[0], 1, dtype=np.int8), ["NotGrouped"])  # This was playa.utils.onescol
    accum = Accum2(cat, temp_cat)

    for (val, name, filt, func) in zip(val_list, name_list, filt_list, func_list):
        func_name = func
        func = getattr(accum, func_name)
        if isinstance(val, list):  # Special cases
            if isinstance(val[1], str):  # Named cases
                if val[1] in "pP":  # accum_ratiop type
                    curr_data = accum_ratiop(cat, temp_cat, val[0], filt, func_name, "T", False, False)
                else:
                    raise ValueError(f'Invalid accum_cols specifier "{val[1]}" in second argument for column {name}')
            else:  # accum_ratio type
                curr_data = accum_ratio(
                    cat, temp_cat, val[0], val[1], filt, filt, func_name, func_name, remove_blanks=False
                )
        else:
            curr_data = func(val, filter=filt)
        try:
            results[name] = curr_data["NotGrouped"]
        except NameError:
            # Get number of keys in (potentially) multikey categorical. This only happens once.
            cat_width = len(cat.category_dict)
            results = curr_data[:, 0:cat_width]
            results.footer_remove()
            results[name] = curr_data["NotGrouped"]
        footer_val = list(curr_data.footer_get_dict().values())[0].get("NotGrouped", 0.0)
        results.footer_set_values("Total", {name: footer_val})

    if remove_blanks:
        return results.trim()
    else:
        return results


# keep this as the last line
TypeRegister.AccumTable = AccumTable



# -*- coding: utf-8 -*-
from itertools import product
from typing import Generator, List, Optional

import numpy as np

from ..rt_fastarray import FastArray
from ..rt_multiset import Multiset
from ..rt_numpy import (
    all,
    any,
    arange,
    argmax,
    argmin,
    argsort,
    isnan,
    lexsort,
    max,
    mean,
    median,
    min,
    minimum,
    nanargmax,
    nanargmin,
    nanmax,
    nanmean,
    nanmedian,
    nanmin,
    nanpercentile,
    nanstd,
    nansum,
    nanvar,
    percentile,
    std,
    sum,
    trunc,
    var,
)
from .benchmark import benchmark
from .runner import benchmark as bench
from .runner import create_comparison_dataset

DEFAULT_ARRAY_SIZES: List[int] = [100, 1000, 10_000, 100_000, 1_000_000, 10_000_000]
DEFAULT_DTYPES: List[type] = [np.int32, np.float32]
DEFAULT_THREADS: List[int] = [1, 2, 4, 8]
DEFAULT_RECYCLER: List[bool] = [True, False]
DEFAULT_REDUCE_FUNCS: List[callable] = [np.sum, np.min, np.std]
DEFAULT_BINARY_FUNCS: List[callable] = [np.add, np.minimum, np.true_divide]
DEFAULT_UNARY_FUNCS: List[callable] = [np.isnan, np.trunc, np.sqrt]


def _gen_arange(
    asizes: Optional[List[int]] = None,
    dtypes: Optional[List[type]] = None,
    nan_indexes: Optional[List[int]] = None,
) -> Generator[np.ndarray, None, None]:
    """_gen_arange is a generator that yields an arrays of cross product of
    array sizes, array data types, and possibly nan-specified indexes.
    The array data is monotonically increasing."""
    if asizes is None:
        asizes = DEFAULT_ARRAY_SIZES
    if dtypes is None:
        dtypes = DEFAULT_DTYPES
    if nan_indexes is None:
        nan_indexes = [None]
    for asize, dtype, nan_index in product(asizes, dtypes, nan_indexes):
        arr = np.arange(asize, dtype=dtype)
        if nan_index is not None and np.dtype(dtype).type in [np.float32, np.float64]:
            arr[nan_index] = np.nan
        yield arr


def gen_arange(
    asizes: Optional[List[int]] = None,
    dtypes: Optional[List[type]] = None,
    nan_indexes: Optional[List[int]] = None,
) -> Generator[np.ndarray, None, None]:
    """gen_arange is a generator of numpy arrays composed of the cross product of
    array sizes, array data types, and possibly nan-specified indexes.
    The array data is monotonically increasing."""
    for arr in _gen_arange(asizes, dtypes, nan_indexes):
        yield arr


def gen_farange(
    asizes: Optional[List[int]] = None,
    dtypes: Optional[List[type]] = None,
    nan_indexes: Optional[List[int]] = None,
) -> Generator[np.ndarray, None, None]:
    """gen_farange is a generator of riptable FastArray composed of the cross product of
    array sizes, array data types, and possibly nan-specified indexes.
    The array data is monotonically increasing."""
    for arr in _gen_arange(asizes, dtypes, nan_indexes):
        yield arr.view(FastArray)


def gen_arange_reverse(
    asizes: Optional[List[int]] = None,
    dtypes: Optional[List[type]] = None,
    nan_indexes: Optional[List[int]] = None,
) -> Generator[np.ndarray, None, None]:
    """A generator of numpy.flip of ``gen_arange``."""
    for arr in _gen_arange(asizes, dtypes, nan_indexes):
        yield np.flip(arr)


def gen_farange_reverse(
    asizes: Optional[List[int]] = None,
    dtypes: Optional[List[type]] = None,
    nan_indexes: Optional[List[int]] = None,
) -> Generator[np.ndarray, None, None]:
    """A generator of numpy.flip of ``gen_farange``."""
    for arr in gen_arange(asizes, dtypes, nan_indexes):
        yield np.flip(arr).view(FastArray)


# ----------------------------------------------------
def benchmark_reduce(
    funcs=DEFAULT_REDUCE_FUNCS,
    dtypes=DEFAULT_DTYPES,
    asizes=DEFAULT_ARRAY_SIZES,
    threads=DEFAULT_THREADS,
    loops=10,
    ratio=True,
    recycle=True,
    scalar=False,
):
    return benchmark(
        funcs=funcs,
        dtypes=dtypes,
        asizes=asizes,
        threads=threads,
        loops=loops,
        ratio=ratio,
        recycle=recycle,
        funcargs=1,
        scalar=scalar,
    )


# ----------------------------------------------------
def benchmark_binary(
    funcs=DEFAULT_BINARY_FUNCS,
    dtypes=DEFAULT_DTYPES,
    asizes=DEFAULT_ARRAY_SIZES,
    threads=DEFAULT_THREADS,
    loops=10,
    ratio=True,
    recycle=True,
    scalar=False,
):
    return benchmark(
        funcs=funcs,
        dtypes=dtypes,
        asizes=asizes,
        threads=threads,
        loops=loops,
        ratio=ratio,
        recycle=recycle,
        funcargs=2,
        scalar=scalar,
    )


# ----------------------------------------------------
def benchmark_unary(
    funcs=DEFAULT_UNARY_FUNCS,
    dtypes=DEFAULT_DTYPES,
    asizes=DEFAULT_ARRAY_SIZES,
    threads=DEFAULT_THREADS,
    loops=10,
    ratio=True,
    recycle=True,
    scalar=False,
):

    return benchmark(
        funcs=funcs,
        dtypes=dtypes,
        asizes=asizes,
        threads=threads,
        loops=loops,
        ratio=ratio,
        recycle=recycle,
        funcargs=1,
        scalar=scalar,
    )


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_all(farr):
    all(farr)


# ----------------------------------------------------
@bench(benchmark_params={"arr": gen_arange()})
def bench_np_all(arr):
    np.all(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(dtypes=[np.float32, np.float64], nan_indexes=[0, -1]),
        "thread_count": DEFAULT_THREADS,
    }
)
def bench_rt_all_with_nan(farr):
    all(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "arr": gen_arange(dtypes=[np.float32, np.float64], nan_indexes=[0, -1]),
    }
)
def bench_np_all_with_nan(arr):
    np.all(arr)


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_any(farr):
    any(farr)


# ----------------------------------------------------
@bench(benchmark_params={"arr": gen_arange()})
def bench_np_any(arr):
    np.any(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(dtypes=[np.float32, np.float64], nan_indexes=[0, -1]),
        "thread_count": DEFAULT_THREADS,
    }
)
def bench_rt_any_with_nan(farr):
    any(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "arr": gen_arange(dtypes=[np.float32, np.float64], nan_indexes=[0, -1]),
    }
)
def bench_np_any_with_nan(arr):
    np.any(arr)


# ----------------------------------------------------
# 20200406 rt.arange punts to np.arange - no need to compare benchmarks
@bench(benchmark_params={"sz": DEFAULT_ARRAY_SIZES, "thread_count": DEFAULT_THREADS})
def bench_rt_arange(sz):
    arange(sz)


# ----------------------------------------------------
@bench(benchmark_params={"sz": DEFAULT_ARRAY_SIZES})
def bench_np_arange(sz):
    np.arange(sz)


# ----------------------------------------------------
# 20200406 rt.argsort punts to np.argsort - no need to compare benchmarks
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "kind": ["quicksort", "mergesort", "heapsort", "stable"],
        "thread_count": DEFAULT_THREADS,
    }
)
def bench_rt_argsort(farr, kind):
    argsort(farr, kind=kind)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "arr": gen_arange(),
        "kind": ["quicksort", "mergesort", "heapsort", "stable"],
    }
)
def bench_np_argsort(arr, kind):
    np.argsort(arr, kind)


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_argmax(farr):
    argmax(farr)


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_nanargmax(farr):
    nanargmax(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "arr": gen_arange(),
    }
)
def bench_np_argmax(arr):
    np.argmax(arr)


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_argmin(farr):
    argmin(farr)


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_nanargmin(farr):
    nanargmin(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "arr": gen_arange(),
    }
)
def bench_np_argmin(arr):
    np.argmin(arr)


# ----------------------------------------------------
# TODO different data types and different ordering (e.g., random, reversed, ordered)
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_lexsort_ordered(farr):
    lexsort((farr, farr))


# ----------------------------------------------------
@bench(benchmark_params={"arr": gen_arange_reverse()})
def bench_np_lexsort_ordered(arr):
    np.lexsort((arr, arr))


# ----------------------------------------------------
@bench(benchmark_params={"farr": gen_farange(), "thread_count": DEFAULT_THREADS})
def bench_rt_lexsort_reversed(farr):
    lexsort((farr, farr))


# ----------------------------------------------------
@bench(benchmark_params={"arr": gen_arange_reverse()})
def bench_np_lexsort_reversed(arr):
    np.lexsort((arr, arr))


# ----------------------------------------------------
# TODO vary percentile and interpolation
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_percentile(farr):
    percentile(farr, 50)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_nanpercentile(farr):
    nanpercentile(farr, 50)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_percentile(arr):
    np.percentile(arr, 50)


# ----------------------------------------------------
@bench(
    benchmark_params={
        # benchmark function parameters
        "farr": gen_farange(),
        # riptable settings
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
    # other benchmark paramters
    benchmark_iterations=5,
)
def bench_rt_sum(farr):
    sum(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_nansum(farr):
    nansum(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
    benchmark_iterations=5,
)
def bench_np_sum(arr):
    np.sum(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_max(farr):
    max(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_nanmax(farr):
    nanmax(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_max(arr):
    np.max(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_mean(farr):
    mean(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_nanmean(farr):
    nanmean(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_mean(arr):
    np.mean(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_median(farr):
    median(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_nanmedian(farr):
    nanmedian(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_median(arr):
    np.median(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_min(farr):
    min(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_nanmin(farr):
    nanmin(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_min(arr):
    np.min(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_std(farr):
    std(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_nanstd(farr):
    nanstd(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_std(arr):
    np.std(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_add(farr):
    # N.B, 20200401 - riptable does not have an implementation for `add` so use the Numpy
    # implementation with FastArray arguments.
    np.add(farr, farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_add(arr):
    np.add(arr, arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_minimum(farr):
    minimum(farr, farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_minimum(arr):
    np.minimum(arr, arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_true_divide(farr):
    # N.B, 20200401 - riptable does not have an implementation for `true_divide` so use the Numpy
    # implementation with FastArray arguments.
    np.true_divide(farr, farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_true_divide(arr):
    np.true_divide(arr, arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_isnan(farr):
    isnan(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_isnan(arr):
    np.isnan(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_trunc(farr):
    trunc(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_trunc(arr):
    np.trunc(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
        "recycler": DEFAULT_RECYCLER,
    },
)
def bench_rt_sqrt(farr):
    # N.B, 20200401 - riptable does not have an implementation for `sqrt` so use the Numpy
    # implementation with FastArray arguments.
    np.sqrt(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_sqrt(arr):
    np.sqrt(arr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_var(farr):
    var(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={
        "farr": gen_farange(),
        "thread_count": DEFAULT_THREADS,
    },
)
def bench_rt_nanvar(farr):
    nanvar(farr)


# ----------------------------------------------------
@bench(
    benchmark_params={"arr": gen_arange()},
)
def bench_np_var(arr):
    np.var(arr)


# ----------------------------------------------------
def compare_argmin():
    return create_comparison_dataset(
        {
            "rt_argmin": bench_rt_argmin(),
            "rt_nanargmin": bench_rt_nanargmin(),
            "np_argmin": bench_np_argmin(),
        }
    )


# ----------------------------------------------------
def compare_argmax():
    return create_comparison_dataset(
        {
            "rt_argmax": bench_rt_argmax(),
            "rt_nanargmax": bench_rt_nanargmax(),
            "np_argmax": bench_np_argmax(),
        }
    )


# ----------------------------------------------------
def compare_all():
    return create_comparison_dataset(
        {
            "rt_all": bench_rt_all(),
            "np_all": bench_np_all(),
        }
    )


# ----------------------------------------------------
def compare_any():
    return create_comparison_dataset(
        {
            "rt_any": bench_rt_any(),
            "np_any": bench_np_any(),
        }
    )


# ----------------------------------------------------
def compare_all_with_nan():
    return create_comparison_dataset(
        {
            "rt_all": bench_rt_all_with_nan(),
            "np_all": bench_np_all_with_nan(),
        }
    )


# ----------------------------------------------------
def compare_any_with_nan():
    return create_comparison_dataset(
        {
            "rt_any": bench_rt_any_with_nan(),
            "np_any": bench_np_any_with_nan(),
        }
    )


# ----------------------------------------------------
def compare_lexsort():
    return create_comparison_dataset(
        {
            "rt_lexsort_ordered": bench_rt_lexsort_ordered(),
            "np_lexsort_ordered": bench_np_lexsort_ordered(),
            "rt_lexsort_reversed": bench_rt_lexsort_reversed(),
            "np_lexsort_reversed": bench_np_lexsort_reversed(),
        }
    )


# ----------------------------------------------------
def compare_max():
    return create_comparison_dataset(
        {
            "rt_max": bench_rt_max(),
            "rt_nanmax": bench_rt_nanmax(),
            "np_max": bench_np_max(),
        }
    )


# ----------------------------------------------------
def compare_mean():
    return create_comparison_dataset(
        {
            "rt_min": bench_rt_mean(),
            "rt_nanmin": bench_rt_nanmean(),
            "np_min": bench_np_mean(),
        }
    )


# ----------------------------------------------------
def compare_median():
    return create_comparison_dataset(
        {
            "rt_median": bench_rt_median(),
            "rt_nanmedian": bench_rt_nanmedian(),
            "np_median": bench_np_median(),
        }
    )


# ----------------------------------------------------
def compare_min():
    return create_comparison_dataset(
        {
            "rt_min": bench_rt_min(),
            "rt_nanmin": bench_rt_nanmin(),
            "np_min": bench_np_min(),
        }
    )


# ----------------------------------------------------
def compare_percentile():
    return create_comparison_dataset(
        {
            "rt_percentile": bench_rt_percentile(),
            "rt_nanpercentile": bench_rt_nanpercentile(),
            "np_percentile": bench_np_percentile(),
        }
    )


# ----------------------------------------------------
def compare_std():
    return create_comparison_dataset(
        {
            "rt_std": bench_rt_std(),
            "rt_nanstd": bench_rt_nanstd(),
            "np_std": bench_np_std(),
        }
    )


# ----------------------------------------------------
def compare_sum():
    return create_comparison_dataset(
        {
            "rt_sum": bench_rt_sum(),
            "rt_nansum": bench_rt_nansum(),
            "np_sum": bench_np_sum(),
        }
    )


# ----------------------------------------------------
def compare_add():
    return create_comparison_dataset(
        {
            "rt_add": bench_rt_add(),
            "np_add": bench_np_add(),
        }
    )


# ----------------------------------------------------
def compare_minimum():
    return create_comparison_dataset(
        {
            "rt_minimum": bench_rt_minimum(),
            "np_minimum": bench_np_minimum(),
        }
    )


# ----------------------------------------------------
def compare_true_divide():
    return create_comparison_dataset(
        {
            "rt_true_divide": bench_rt_true_divide(),
            "np_true_divide": bench_np_true_divide(),
        }
    )


# ----------------------------------------------------
def compare_isnan():
    return create_comparison_dataset(
        {
            "rt_isnan": bench_rt_isnan(),
            "np_isnan": bench_np_isnan(),
        }
    )


# ----------------------------------------------------
def compare_trunc():
    return create_comparison_dataset(
        {
            "rt_trunc": bench_rt_trunc(),
            "np_trunc": bench_np_trunc(),
        }
    )


# ----------------------------------------------------
def compare_sqrt():
    return create_comparison_dataset(
        {
            "rt_sqrt": bench_rt_sqrt(),
            "np_sqrt": bench_np_sqrt(),
        }
    )


# ----------------------------------------------------
def compare_var():
    return create_comparison_dataset(
        {
            "rt_var": bench_rt_var(),
            "rt_nanvar": bench_rt_nanvar(),
            "np_var": bench_np_var(),
        }
    )


# ----------------------------------------------------
def compare_numpy_reduce():
    return Multiset(
        {
            "min": compare_min(),
            "std": compare_std(),
            "sum": compare_sum(),
        }
    )


# ----------------------------------------------------
def compare_numpy_binary():
    return Multiset(
        {
            "add": compare_add(),
            "minimum": compare_minimum(),
            "true_divide": compare_true_divide(),
        }
    )


# ----------------------------------------------------
def compare_numpy_unary():
    return Multiset(
        {
            "isnan": compare_isnan(),
            "trunc": compare_trunc(),
            "sqrt": compare_sqrt(),
        }
    )



__all__ = ["MetaData", "meta_from_version", "META_VERSION"]

import json

from ..rt_enum import CategoryMode, DisplayLength, TypeId, TypeRegister

# global meta version, saved starting 12/17/2018 - anything prior will default to 0
META_VERSION = 0

version_dict = {
    # this dictionary will be checked whwnever a riptable class is being rebuilt during an SDS load
    # top level keys are the version number
    # next keys are the class's type id - see TypeId in enum
    0: {
        TypeId.Categorical: {
            # vars for container loader
            "name": "Categorical",
            "typeid": TypeId.Categorical,
            "version": 0,
            # vars for additional arrays
            "colnames": [],
            "ncols": 0,
            # vars to rebuild the same categorical
            "instance_vars": {
                "mode": CategoryMode.StringArray,
                "base_index": 1,
                "ordered": False,
                "sorted": False,
            },
        },
        TypeId.Dataset: None,
        TypeId.Struct: None,
        TypeId.DateTimeNano: {
            "name": "DateTimeNano",
            "typeid": TypeId.DateTimeNano,
            "ncols": 0,
            "version": 0,
            "instance_vars": {
                "_display_length": DisplayLength.Long,
                "_timezone_str": "America/New York",
                "_to_tz": "NYC",
            },
        },
        TypeId.TimeSpan: {
            "name": "TimeSpan",
            "typeid": TypeId.TimeSpan,
            "ncols": 0,
            "version": 0,
            "instance_vars": {"_display_length": DisplayLength.Long},
        },
    }
}


def meta_from_version(cls, vnum):
    """
    Returns a dictionary of meta data defaults.
    """
    id = getattr(TypeId, cls.__name__)
    return version_dict[vnum][id]


class MetaData:

    default_dict = {"name": "", "typeid": TypeId.Default}

    def __init__(self, metadict={}):
        self._dict = self.default_dict.copy()

        if isinstance(metadict, MetaData):
            self._dict = metadict._dict

        else:
            if isinstance(metadict, (bytes, str)):
                metadict = json.loads(metadict)

            for k, v in metadict.items():
                self._dict[k] = v

    @property
    def string(self):
        return json.dumps(self._dict)

    @property
    def dict(self):
        return self._dict

    @property
    def name(self):
        return self["name"]

    @property
    def typeid(self):
        return self["typeid"]

    @property
    def itemclass(self):
        """Starting 4/29/2019 item classes will be saved as strings in json meta data in `classname`.
        For backwards compatibility, will also check `typeid`. The TypeId class is an enum of ItemClass -> typeid.
        Both will lookup the items class in the TypeRegister, which holds classname -> itemclass.
        """
        try:
            classname = self["classname"]
        except:
            classname = TypeId(self["typeid"]).name
        return getattr(TypeRegister, classname)

    # pass these to dict
    # ------------------------------------------------------------
    def __getitem__(self, idx):
        return self._dict[idx]

    def __setitem__(self, idx, value):
        self._dict[idx] = value

    def get(self, key, default):
        return self._dict.get(key, default)

    def setdefault(self, k, v):
        self._dict.setdefault(k, v)

    def __repr__(self):
        return self._dict.__repr__()

    def __str__(self):
        return self._dict.__str__()

    # ------------------------------------------------------------



import string
import unittest
from typing import Any, List, Optional, Union

import riptable as rt
from riptable.rt_enum import CategoryMode

#
# TODO: Consider moving these functions into a 'testing' submodule of riptable, a la numpy.testing.
#


_N = 10

# The following helper methods are used to generate the underlying data a Categorical of the various
# category modes will use when constructed.


def get_categorical_base_data() -> List[List[str]]:
    # This acts a a base for the different edge cases considered for the type of data that is used when constructing
    # a Categorical. The other Categorical data functions, such as the one for bytes and numerics, derive from this table.
    # If new edge cases are considered, this table can be extended which will extend tables for the other data types.
    return [
        # [],  # empty list; commented out since this case throws a ValueError in Categorical constructor.
        ["a"],  # one element list
        ["a"] * _N,  # list of single repeated element
        list(string.ascii_letters),  # list of all unique elements
        ["a"] * _N + ["z"] + ["a"] * _N,  # repeats surrounding single unique
        ["a"] + ["b"] * _N + ["c"],  # single unique surrounded by repeats
    ]


def _get_categorical_byte_data() -> List[List[bytes]]:
    bytes = []
    for values in get_categorical_base_data():
        b = [s.encode("utf-8") for s in values]
        bytes.extend(b)
    return bytes


def _get_categorical_numeric_data() -> List[List[int]]:
    numerics = []
    for values in get_categorical_base_data():
        numerics.append([ord(c) for c in values])
    return numerics


def _get_categorical_multikey_data() -> List[List[rt.FastArray]]:
    # Create data that can be used to construct MultiKey Categoricals where both keys are
    # strings, numerics, and combination of the two.
    strings = get_categorical_base_data()
    numerics = _get_categorical_numeric_data()

    results = []
    # consider parameterizing over the number of keys instead of literal handling of up to four keys
    for values in strings + numerics:  # two keys of same dtype and value
        results.append([rt.FA(values), rt.FA(values)])
    for values, values1 in zip(strings, numerics):  # two keys of different dtypes
        results.append([rt.FA(values), rt.FA(values1)])
    for values, values1, values2 in zip(strings, strings, numerics):  # three keys of different dtypes
        results.append([rt.FA(values), rt.FA(values1), rt.FA(values2)])
    for values, values1, values2, values3 in zip(strings, strings, strings, numerics):  # four keys of different dtypes
        results.append([rt.FA(values), rt.FA(values1), rt.FA(values2), rt.FA(values3)])
    return results


def verbose_categorical(categorical: rt.Categorical) -> str:
    lst = []
    lst.append(f"Categorical:\n{repr(categorical)}")
    lst.append(f"Type:\n{type(categorical)}")
    lst.append(f"Category mode:\n{repr(rt.rt_enum.CategoryMode(categorical.category_mode))}")
    lst.append(f"Lock:\n{categorical._locked}")
    # add support for multikey categoricals
    if categorical.category_mode != rt.rt_enum.CategoryMode.MultiKey:
        lst.append(f"_fa\n{repr(categorical._fa)}")
        lst.append(f"category_array\n{repr(categorical.category_array)}")
        lst.append(f"category_dict\n{categorical.category_dict}")
        if categorical.category_mode == rt.rt_enum.CategoryMode.Dictionary:
            lst.append(f"category_mapping\n{categorical.category_mapping}")
            lst.append(f"category_codes\n{categorical.category_codes}")
    return "\n\n".join(lst)


def get_categorical_data_factory_method(
    category_modes: Optional[Union[CategoryMode, List[CategoryMode]]] = None
) -> List[List[Any]]:
    """Get test data used to construct Categoricals.

    Parameters
    ----------
    category_modes : CategoryMode, list of CategoryMode, optional
        Category modes that are used for getting the appropriate data to construct a Categorical. If no category_mode is specified, data for all supported CategoryModes will be returned, otherwise select the type of data by specifying the CategoryMode or a list of CategoryModes.

    Returns
    -------
    data : List[List[Any]]
        - The list of values that can be used to construct a Categorical.

    Examples
    --------
    >>> data = get_categorical_data_factory_method([CategoryMode.StringArray, CategoryMode.NumericArray])
    >>> cats = [Categorical(values) for values in data]
    """
    if category_modes is None:  # return all types of categories
        return get_categorical_base_data() + _get_categorical_numeric_data() + _get_categorical_multikey_data()
    if not isinstance(category_modes, list):  # wrap single category mode in a list
        category_modes = [category_modes]

    underlying_data = []
    if rt.rt_enum.CategoryMode.StringArray in category_modes:
        underlying_data.extend(get_categorical_base_data())
    if rt.rt_enum.CategoryMode.NumericArray in category_modes:
        underlying_data.extend(_get_categorical_numeric_data())
    if rt.rt_enum.CategoryMode.MultiKey in category_modes:
        underlying_data.extend(_get_categorical_multikey_data())
    return underlying_data


def get_all_categorical_data() -> List[rt.Categorical]:
    """Returns a list of all the Categorical test data of all supported CategoryModes."""
    return [rt.Categorical(data) for data in get_categorical_data_factory_method()]


if __name__ == "__main__":
    tester = unittest.main()
